from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict, List, Optional, Union

import attr

from blockapi.utils.datetime import parse_dt
from blockapi.utils.num import raw_to_decimals, to_decimal, to_int

UNKNOWN = 'unknown'


class Blockchain(str, Enum):
    ACALA = 'acala'
    AKASH = 'akash'
    ALGORAND = 'algorand'
    ARBITRUM = 'arbitrum'
    ARBITRUM_NOVA = 'arbitrum-nova'
    ARBITRUM_ONE = 'arbitrum-one'
    ARDOR = 'ardor'
    ASTAR = 'astar'
    AURORA = 'aurora'
    AVALANCHE = 'avalanche'
    AXIE = 'axie'
    BINANCECOIN = 'binancecoin'
    BINANCE_SMART_CHAIN = 'binance-smart-chain'
    BITCICHAIN = 'bitcichain'
    BITCOIN = 'bitcoin'
    BITCOIN_CASH = 'bitcoin-cash'
    BITGERT = 'bitgert'
    BITSHARES = 'bitshares'
    BIT_TORRENT = 'bit-torrent'
    BLOCKNET = 'blocknet'
    BNB_BEACON_CHAIN = 'bnb-beacon-chain'
    BOBA = 'boba'
    BOS = 'bos'
    CANTO = 'canto'
    CARDANO = 'cardano'
    CELER_NETWORK = 'celer-network'
    CELO = 'celo'
    CHILIZ = 'chiliz'
    CLV_PARACHAIN = 'clv-parachain'
    CMP = 'cmp'
    COINEX_SMART_CHAIN = 'coinex-smart-chain'
    COLOSSUSXT = 'colossusxt'
    CONFLUX = 'conflux'
    COSMOS = 'cosmos'
    COUNTERPARTY = 'counterparty'
    CRONOS = 'cronos'
    CRYPTO_ORG = 'crypto-org'
    CUBE = 'cube'
    DARWINIA_CRAB_NETWORK = 'darwinia-crab-network'
    DASH = 'dash'
    DEFI_KINGDOMS = 'defi-kingdoms'
    DFK_CHAIN = 'dfk-chain'
    DITHEREUM = 'dithereum'
    DOGECHAIN = 'dogechain'
    DXCHAIN = 'dxchain'
    ECHELON = 'echelon'
    ELASTOS = 'elastos'
    ELA_DID_SIDECHAIN = 'ela-did-sidechain'
    ELROND = 'elrond'
    EMERALD_PARATIME = 'emerald-paratime'
    ENERGI = 'energi'
    ENQ_ENECUUM = 'enq-enecuum'
    EOS = 'eos'
    ETHEREUM = 'ethereum'
    ETHEREUMPOW = 'ethereumpow'
    ETHEREUM_CLASSIC = 'ethereum-classic'
    EVERSCALE = 'everscale'
    EVMOS = 'evmos'
    EXPANSE_NETWORK = 'expanse-network'
    FACTOM = 'factom'
    FANTOM = 'fantom'
    FINDORA = 'findora'
    FLARE = 'flare'
    FUSE = 'fuse'
    FUSION_NETWORK = 'fusion-network'
    GENESIS_L1 = 'genesis-l1'
    GNOSIS = 'gnosis'
    GOCHAIN = 'gochain'
    GODWOKEN = 'godwoken'
    HARMONY = 'harmony'
    HECO = 'heco'
    HEDERA_HASHGRAPH = 'hedera-hashgraph'
    HOO_SMART_CHAIN = 'hoo-smart-chain'
    HUOBI_TOKEN = 'huobi-token'
    HYDRA = 'hydra'
    ICON = 'icon'
    IOTEX = 'iotex'
    IRIS = 'iris'
    JUNO = 'juno'
    KARDIACHAIN = 'kardiachain'
    KARURA = 'karura'
    KAVA = 'kava'
    KCC = 'kcc'
    KLAYTN_CYPRESS = 'klaytn-cypress'
    KLAY_TOKEN = 'klay-token'
    KOMODO = 'komodo'
    KUCOIN = 'kucoin'
    KUSAMA = 'kusama'
    LITECOIN = 'litecoin'
    METAVERSE_ETP = 'metaverse-etp'
    METER = 'meter'
    METIS_ANDROMEDA = 'metis-andromeda'
    MILKOMEDA_C1 = 'milkomeda-c1'
    MILKOMEDA_CARDANO = 'milkomeda-cardano'
    MIXIN_NETWORK = 'mixin-network'
    MOONBEAM = 'moonbeam'
    MOONBEAM_MOONRIVER = 'moonbeam-moonriver'
    MOONRIVER = 'moonriver'
    NEAR_PROTOCOL = 'near-protocol'
    NEM = 'nem'
    NEO = 'neo'
    NULS = 'nuls'
    NXT = 'nxt'
    OASIS = 'oasis'
    OKEX_CHAIN = 'okex-chain'
    OKT = 'okt'
    OMNI = 'omni'
    ONTOLOGY = 'ontology'
    OPENLEDGER = 'openledger'
    OPTIMISM = 'optimism'
    OPTIMISTIC_ETHEREUM = 'optimistic-ethereum'
    OSMOSIS = 'osmosis'
    PALM = 'palm'
    PERSISTENCE = 'persistence'
    PLAT_ON = 'plat-on'
    POLIS_CHAIN = 'polis-chain'
    POLKADOT = 'polkadot'
    POLYGON = 'polygon'
    QTUM = 'qtum'
    REGEN = 'regen'
    REI_NETWORK = 'rei-network'
    RONIN = 'ronin'
    ROOTSTOCK = 'rootstock'
    SECRET = 'secret'
    SEED_COIN_NETWORK = 'seed-coin-network'
    SENTINEL = 'sentinel'
    SHIBA_CHAIN = 'shiba-chain'
    SHIDEN_NETWORK = 'shiden-network'
    SIFCHAIN = 'sifchain'
    SKALE = 'skale'
    SMART_BITCOIN_CASH = 'smart-bitcoin-cash'
    SOLANA = 'solana'
    SONGBIRD = 'songbird'
    SONGBIRD_CANARY = 'songbird-canary'
    SORA = 'sora'
    STACKS = 'stacks'
    STELLAR = 'stellar'
    STEP_NETWORK = 'step-network'
    STRATIS = 'stratis'
    SUPER_ZERO = 'super-zero'
    SX_NETWORK = 'sx-network'
    SYSCOIN = 'syscoin'
    TBWG_CHAIN = 'tbwg-chain'
    TELOS = 'telos'
    TERRA = 'terra'
    TEZOS = 'tezos'
    THAI_CHAIN = 'thai-chain'
    THAI_CHAIN_2_THAI_FI = 'thai-chain-2-thai-fi'
    THETA = 'theta'
    THOR = 'thor'
    THUNDERCORE = 'thundercore'
    TOMBCHAIN = 'tombchain'
    TOMOCHAIN = 'tomochain'
    TRON = 'tron'
    UBIQ = 'ubiq'
    VALORBIT = 'valorbit'
    VECHAIN = 'vechain'
    VELAS = 'velas'
    VITE = 'vite'
    WANCHAIN = 'wanchain'
    WAVES = 'waves'
    XDAI = 'xdai'
    XDC_NETWORK = 'xdc-network'
    XRP = 'xrp'
    YOCOIN = 'yocoin'
    ZILLIQA = 'zilliqa'


class AssetType(str, Enum):
    AVAILABLE = 'available'
    CLAIMABLE = 'claimable'
    COLLATERAL = 'collateral'
    COMMON = 'common'
    DEBT = 'debt'
    DEPOSITED = 'deposited'
    FARMING = 'farming'
    INVESTMENT = 'investment'
    LENDING = 'lending'
    LENDING_BORROW = 'lending_borrow'
    LENDING_REWARD = 'lending_reward'
    LIQUIDITY_POOL = 'liquidity_pool'
    LOCKED = 'locked'
    REWARDS = 'rewards'
    STAKED = 'staked'
    UNREALIZED_AVAILABLE = 'unrealized_available'
    VESTING = 'vesting'
    YIELD = 'yield'

    # DEPRECATED
    ASSET = 'asset'


class TokenRole(str, Enum):
    SUPPLY = 'supply'
    REWARD = 'reward'
    BORROW = 'borrow'
    LIQUIDITY_POOL = 'liquidity_pool'


class OperationType(str, Enum):
    UNKNOWN = 'unknown'
    INFLATION = 'inflation'
    TRANSACTION = 'transaction'
    COLLECT_TX_FEE = 'collect-tx-fee'


class OperationDirection(str, Enum):
    OUTGOING = 'outgoing'
    INCOMING = 'incoming'


class TransactionStatus(str, Enum):
    CONFIRMED = 'confirmed'
    PENDING = 'pending'


@attr.s(auto_attribs=True, slots=True)
class ApiOptions:
    blockchain: Blockchain
    base_url: str
    rate_limit: float = 0.0
    testnet: bool = False

    start_offset: Optional[int] = attr.ib(default=None)
    max_items_per_page: Optional[int] = attr.ib(default=None)
    page_offset_step: Optional[int] = attr.ib(default=None)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class CoinInfo:
    tags: List[str] = attr.ib(default=None)
    total_supply: Optional[Decimal] = attr.ib(default=None)
    logo_url: Optional[str] = attr.ib(default=None)
    coingecko_id: Optional[str] = attr.ib(default=None)
    website: Optional[str] = attr.ib(default=None)

    @classmethod
    def from_api(
        cls,
        tags: Optional[List[str]] = None,
        total_supply: Optional[Union[int, float, str]] = None,
        logo_url: Optional[str] = None,
        coingecko_id: Optional[str] = None,
        website: Optional[str] = None,
    ) -> 'CoinInfo':
        return cls(
            tags=tags,
            total_supply=(
                to_decimal(total_supply) if total_supply is not None else None
            ),
            logo_url=logo_url,
            coingecko_id=coingecko_id,
            website=website,
        )


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Coin:
    symbol: str
    name: str
    decimals: int
    blockchain: Blockchain
    address: Optional[str] = attr.ib(default=None)
    standards: Optional[List[str]] = attr.ib(default=None)
    protocol_id: Optional[str] = attr.ib(default=None)
    info: Optional[CoinInfo] = attr.ib(default=None)

    @classmethod
    def from_api(
        cls,
        blockchain: Blockchain,
        decimals: Union[int, str],
        symbol: Optional[str] = None,
        name: Optional[str] = None,
        address: Optional[str] = None,
        standards: Optional[List[str]] = None,
        protocol_id: Optional[str] = None,
        info: Optional[CoinInfo] = None,
    ) -> 'Coin':
        return cls(
            symbol=symbol if symbol is not None else UNKNOWN,
            name=name if name is not None else UNKNOWN,
            decimals=to_int(decimals) if decimals is not None else 0,
            blockchain=blockchain,
            address=address,
            standards=standards,
            protocol_id=protocol_id,
            info=info,
        )


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Protocol:
    protocol_id: str
    chain: Blockchain
    name: str
    user_deposit: Decimal
    site_url: Optional[str] = attr.ib(default=None)
    logo_url: Optional[str] = attr.ib(default=None)
    has_supported_portfolio: bool = attr.ib(default=False)

    @classmethod
    def from_api(
        cls,
        *,
        protocol_id: str,
        chain: Blockchain,
        name: str,
        user_deposit: Union[str, float, int],
        site_url: Optional[str] = None,
        logo_url: Optional[str] = None,
        has_supported_portfolio: Optional[bool] = False,
    ) -> 'Protocol':
        return cls(
            protocol_id=protocol_id,
            chain=chain,
            name=name,
            user_deposit=to_decimal(user_deposit),
            site_url=site_url,
            logo_url=logo_url,
            has_supported_portfolio=has_supported_portfolio,
        )


@attr.s(auto_attribs=True, slots=True, frozen=True)
class BalanceItem:
    balance: Decimal
    balance_raw: Decimal
    raw: Dict
    coin: Coin
    asset_type: AssetType = AssetType.AVAILABLE
    last_updated: Optional[datetime] = attr.ib(default=None)
    protocol: Optional[Protocol] = attr.ib(default=None)
    is_wallet: bool = True
    token_set: Optional[List[str]] = attr.ib(default=None)
    token_role: Optional[TokenRole] = attr.ib(default=None)
    pool_id: Optional[str] = attr.ib(default=None)

    @classmethod
    def from_api(
        cls,
        *,
        balance_raw: Union[int, float, str],
        coin: Coin,
        asset_type: AssetType = AssetType.AVAILABLE,
        raw: Dict,
        last_updated: Optional[Union[int, str]] = None,
        protocol: Optional[Protocol] = None,
        is_wallet: bool = True,
        token_set: Optional[List[str]] = None,
        token_role: Optional[TokenRole] = None,
        pool_id: Optional[str] = None,
    ) -> 'BalanceItem':
        return cls(
            balance_raw=to_decimal(balance_raw),
            balance=raw_to_decimals(balance_raw, coin.decimals),
            coin=coin,
            asset_type=asset_type,
            raw=raw,
            last_updated=(parse_dt(last_updated) if last_updated is not None else None),
            protocol=protocol,
            is_wallet=is_wallet,
            token_set=token_set,
            token_role=token_role,
            pool_id=pool_id,
        )

    def __add__(self, other: 'BalanceItem') -> 'BalanceItem':
        """
        Warning: Adding items of different coins leads to wrong results.
        """
        return BalanceItem(
            balance_raw=self.balance_raw + other.balance_raw,
            balance=self.balance + other.balance,
            coin=self.coin,
            asset_type=self.asset_type,
            raw=self._add_raw(other),
            last_updated=self.last_updated,
            protocol=self.protocol,
            is_wallet=self.is_wallet,
        )

    def _add_raw(self, other):
        """
        Used to skip wrapping of "merged" into "merged" key over and over again.
        """
        if self.raw.get("merged") and isinstance(self.raw.get("merged"), list):
            return {"merged": self.raw.get("merged") + [other.raw]}
        else:
            return {"merged": [self.raw, other.raw]}


@attr.s(auto_attribs=True, slots=True, frozen=True)
class OperationItem:
    amount: Decimal
    amount_raw: Decimal
    coin: Coin
    from_address: str
    to_address: str
    hash: str
    type: OperationType
    direction: OperationDirection
    confirmed: Optional[datetime]
    raw: dict

    @classmethod
    def from_api(
        cls,
        *,
        amount_raw: Union[int, float, str],
        coin: Coin,
        from_address: str,
        to_address: str,
        hash: str,
        type: OperationType,
        direction: OperationDirection,
        raw: dict,
        confirmed: Optional[Union[int, str]] = None,
    ) -> 'OperationItem':
        return cls(
            amount_raw=to_decimal(amount_raw),
            amount=raw_to_decimals(amount_raw, coin.decimals),
            coin=coin,
            from_address=from_address,
            to_address=to_address,
            hash=hash,
            type=type,
            direction=direction,
            confirmed=(parse_dt(confirmed) if confirmed is not None else None),
            raw=raw,
        )


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TransactionItem:
    fee: Decimal
    fee_raw: Decimal
    date: datetime
    coin: Coin
    hash: str
    status: TransactionStatus
    operations: List[OperationItem]
    raw: dict

    @classmethod
    def from_api(
        cls,
        *,
        fee_raw: Union[int, float, str],
        coin: Coin,
        date: Union[int, str],
        hash: str,
        status: TransactionStatus = TransactionStatus.CONFIRMED,
        operations: List[OperationItem],
        raw: dict,
    ) -> 'TransactionItem':
        return cls(
            fee_raw=to_decimal(fee_raw or '0'),
            fee=raw_to_decimals(fee_raw or '0', coin.decimals),
            coin=coin,
            date=(parse_dt(date) if date is not None else None),
            hash=hash,
            status=status,
            operations=operations,
            raw=raw,
        )


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Pool:
    pool_id: str
    protocol: Protocol
    items: List[BalanceItem]
    locked_until: Optional[datetime] = attr.ib(default=None)
    health_rate: Optional[Decimal] = attr.ib(default=None)
    token_set: Optional[str] = attr.ib(default=None)
    project_id: Optional[str] = attr.ib(default=None)
    adapter_id: Optional[str] = attr.ib(default=None)

    @classmethod
    def from_api(
        cls,
        *,
        pool_id: str,
        protocol: Protocol,
        locked_until: Optional[Union[int, str, float]] = None,
        health_rate: Optional[Union[float, str]] = None,
        items: List[BalanceItem],
        token_set: Optional[str] = None,
        project_id: Optional[str] = None,
        adapter_id: Optional[str] = None,
    ) -> 'Pool':
        return cls(
            pool_id=pool_id,
            protocol=protocol,
            items=items,
            locked_until=(parse_dt(locked_until) if locked_until is not None else None),
            health_rate=to_decimal(health_rate) if health_rate is not None else None,
            token_set=token_set,
            project_id=project_id,
            adapter_id=adapter_id,
        )

    def append_items(self, items: List[BalanceItem]) -> None:
        self.items.extend(items)



import json
from functools import lru_cache
from typing import Dict, List, Optional, Sequence

from cytoolz import concatv
from requests import Response

from blockapi.v2.base import (
    ApiException,
    ApiOptions,
    BlockchainApi,
    IBalance,
    InvalidAddressException,
)
from blockapi.v2.coins import COIN_TERRA
from blockapi.v2.models import AssetType, BalanceItem, Blockchain, Coin, CoinInfo


class TerraApi(IBalance):
    """
    Terra Money, implemented by multiple api providers.
    Explorer: https://finder.terra.money
    """

    coin = COIN_TERRA

    def __init__(self):
        self.mantle = TerraMantleApi()
        self.fcd = TerraFcdApi()

    def get_balance(self, address: str) -> List[BalanceItem]:
        native_balances = self.fcd.get_native_balances(address)
        staking_balances = self.fcd.get_staking_balances(address)
        cw20_balances = self.mantle.get_cw20_balances(address)
        return list(concatv(native_balances, staking_balances, cw20_balances))


class TerraFcdApi(BlockchainApi):
    """
    Terra Money FCD
    API docs: https://fcd.terra.dev/swagger
    """

    coin = COIN_TERRA
    api_options = ApiOptions(
        blockchain=Blockchain.TERRA,
        base_url='https://fcd.terra.dev/',
    )

    supported_requests = {
        'get_native_balances': '/v1/bank/{address}',
        'get_ibc_denom_trace': '/ibc/apps/transfer/v1/denom_traces/{hash}',
        'get_staking_data': '/v1/staking/{address}',
    }

    def get_native_balances(self, address: str) -> List[BalanceItem]:
        response = self.get('get_native_balances', address=address)

        balances = []
        for b in response['balance']:
            if int(b['available']) == 0:
                continue

            coin = (
                self._get_terra_token_by_denom(b['denom'])
                if b['denom'].startswith('u')
                else self._get_ibc_token_by_denom(b['denom'])
            )

            balances.append(
                BalanceItem.from_api(balance_raw=b['available'], coin=coin, raw=b)
            )

        return balances

    def get_staking_balances(self, address: str) -> List[BalanceItem]:
        response = self.get('get_staking_data', address=address)

        total_staked = 0
        balances = []

        # active stake
        if int(response['delegationTotal']) > 0:
            total_staked += int(response['delegationTotal'])
        # undelegated stake
        if response['undelegations']:
            total_staked += sum(int(u['amount']) for u in response['undelegations'])
        # total stake - sum of staked and undelegated
        # add redelegations?
        if total_staked:
            balances.append(
                BalanceItem.from_api(
                    balance_raw=total_staked,
                    coin=self.coin,
                    asset_type=AssetType.STAKED,
                    raw=response,
                )
            )

        # staking rewards
        for d in response['rewards']['denoms']:
            balances.append(
                BalanceItem.from_api(
                    balance_raw=d['amount'],
                    coin=self._get_terra_token_by_denom(d['denom']),
                    asset_type=AssetType.CLAIMABLE,
                    raw=d,
                )
            )

        return balances

    # It's possible to get cw20 balances, but it needs to be done one by one.
    # Use .terra_mantle.py for that
    # def get_cw20_balances(self):

    @staticmethod
    def _get_terra_token_by_denom(denom: str) -> Coin:
        if denom == 'uluna':
            return COIN_TERRA
        else:
            symbol = f'{denom[1:3].upper()}TC'
            return Coin.from_api(
                symbol=symbol,
                name=symbol,
                decimals=6,
                blockchain=Blockchain.TERRA,
                address=denom,
                standards=['terra-native'],
            )

    @lru_cache(maxsize=8)
    def _get_ibc_token_by_denom(self, denom: str) -> Coin:
        hash_ = denom.split('/')[1]
        try:
            response = self.get('get_ibc_denom_trace', hash=hash_)
        except ApiException:
            # add log
            symbol = None
        else:
            denom = response['denom_trace']['base_denom']
            symbol = denom[1:].upper()

        return Coin.from_api(
            symbol=symbol,
            name=symbol,
            decimals=6,
            blockchain=Blockchain.TERRA,
            address=hash_,
            standards=['ibc'],
        )


class TerraMantleApi(BlockchainApi):
    """
    Terra Money Subgraph API
    API docs: https://mantle.terra.dev
    """

    coin = COIN_TERRA
    api_options = ApiOptions(
        blockchain=Blockchain.TERRA,
        base_url='https://mantle.terra.dev',
    )

    # API uses post requests
    supported_requests = {}
    _post_requests = {
        'wasm_contract_address_store': """
            WasmContractsContractAddressStore(
                ContractAddress: "$CONTRACT_ADDRESS",
                QueryMsg: "$QUERY_MSG"
            ){
                Result
            }
        """
    }

    _tokens_map: Optional[Dict[str, Dict]] = None

    @property
    def tokens_map(self) -> Dict[str, Dict]:
        if self._tokens_map is None:
            response = self._session.get('https://assets.terra.money/cw20/tokens.json')
            token_list = response.json()
            self._tokens_map = token_list['classic']

        return self._tokens_map

    def get_cw20_balances(self, address: str):
        raw_balances = self._get_raw_balances(address)

        balances = []
        for contract, result_raw in raw_balances['data'].items():
            if not result_raw:
                # should be error in response, TODO add log
                continue

            data_raw = json.loads(result_raw['Result'])
            balance_raw = data_raw['balance']
            if int(balance_raw) == 0:
                continue

            balances.append(
                BalanceItem.from_api(
                    balance_raw=balance_raw,
                    coin=self._get_token_data(contract),
                    raw=result_raw,
                )
            )

        return balances

    def _get_token_data(self, address: str) -> Coin:
        raw_token = self.tokens_map[address]
        return Coin(
            symbol=raw_token['symbol'],
            name=raw_token['name'] if raw_token.get('name') else raw_token['symbol'],
            decimals=6,
            blockchain=Blockchain.TERRA,
            address=address,
            standards=['CW20'],
            protocol_id=raw_token.get('protocol'),
            info=CoinInfo.from_api(logo_url=raw_token.get('icon')),
        )

    def _get_raw_balances(self, address: str) -> Dict:
        cw20_contracts = list(self.tokens_map.keys())
        message = '{\\"balance\\": {\\"address\\": \\"$ADDR\\"}}'.replace(
            '$ADDR', address
        )

        key_queries = [
            self._create_key_query(
                key=contract,
                query=self._build_query(
                    method='wasm_contract_address_store',
                    params={'$CONTRACT_ADDRESS': contract, '$QUERY_MSG': message},
                ),
            )
            for contract in cw20_contracts
        ]
        query = self._concat_key_queries(key_queries)
        return self.post(json={'query': query})

    def _build_query(self, method: str, params: Optional[Dict[str, str]] = None) -> str:
        query = self._post_requests.get(method)
        if params:
            for k, v in params.items():
                query = query.replace(k, v)
        return query

    @staticmethod
    def _create_key_query(key: str, query: str) -> str:
        return f'{key}: {query}'

    @staticmethod
    def _concat_key_queries(key_queries: Sequence[str]) -> str:
        return '{' + ',\n'.join(key_queries) + '}'

    def _opt_raise_on_other_error(self, response: Response) -> None:
        json_response = response.json()
        if json_response.get('errors') is None:
            return

        # pick first message
        err = json_response['errors'][0]

        if 'addr_canonicalize' in err['message']:
            raise InvalidAddressException(f'Invalid address format.')



from os import environ

from pytest import mark

from blockapi.api import AmberdataAPI
from blockapi.test_data import test_addresses


class TestAmberdataAPI:
    ADDRESS = test_addresses['ETH'][0]
    API_KEY = environ.get('AMBERDATAAPI_KEY', 'None')

    @mark.vcr(filter_headers=['x-api-key'])
    def test_get_balance(self):
        api = AmberdataAPI(self.ADDRESS, self.API_KEY)
        result = api.get_balance()

        assert len(result) == 57
        eth_bal = next(b for b in result if b['symbol'] == 'ETH')
        assert eth_bal['amount'] == 325000.4290170703

    @mark.vcr(filter_headers=['x-api-key'])
    def test_get_token_info(self):
        result = AmberdataAPI.get_token_info(
            token_address='0xdac17f958d2ee523a2206206994597c13d831ec7',
            api_key=self.API_KEY,
        )

        assert result['symbol'] == 'USDT'



from decimal import Decimal

import pytest

from blockapi.v2.api.optimistic_etherscan import OptimismEtherscanApi

ETH_BALANCE = '248936789990053052'


@pytest.fixture
def optimism_etherscan_api():
    return OptimismEtherscanApi(api_key="test_key")


@pytest.fixture
def api_response():
    return {'status': '1', 'message': 'OK', 'result': ETH_BALANCE}


def test_parse_eth_balance(optimism_etherscan_api, api_response):
    parsed_balance = optimism_etherscan_api._parse_eth_balance(api_response)
    assert parsed_balance.balance_raw == Decimal(ETH_BALANCE)
    assert parsed_balance.raw == api_response



import json
from abc import ABC
from decimal import Decimal
from typing import Iterable, List, Optional

from requests import Response

from blockapi.utils.num import decimals_to_raw, safe_opt_decimal, to_decimal
from blockapi.v2.base import (
    ApiException,
    ApiOptions,
    BlockchainApi,
    IBalance,
    InvalidAddressException,
)
from blockapi.v2.coins import COIN_DOT, COIN_KSM
from blockapi.v2.models import AssetType, BalanceItem, Blockchain


class SubscanApi(BlockchainApi, IBalance, ABC):
    """
    API docs: https://docs.api.subscan.io/
    Explorer: https://www.subscan.io
    """

    coin = None
    api_options = None

    supported_requests = {
        'get_balance': '/api/v2/scan/search',
        'get_rewards': '/api/v2/scan/account/reward_slash',
    }

    def get_balance(self, address: str) -> List[BalanceItem]:
        balances = []
        balances.extend(list(self._yield_native_balances(address)))

        # add staking rewards (and slashes) too? it's a lot of requests
        # per single address
        # if staking_reward := self._get_staking_reward(address):
        #     balances.append(staking_reward)

        return balances

    def _yield_native_balances(self, address: str) -> Iterable[BalanceItem]:
        body = json.dumps({'key': address})
        response = self._post('get_balance', body=body)

        data = response['data']['account']
        b_total = decimals_to_raw(data['balance'], self.coin.decimals)
        if b_total == 0:
            return []

        b_total_locked = decimals_to_raw(data['balance_lock'], self.coin.decimals)

        b_staked = Decimal(0)
        if data['staking_info']:
            b_staked = b_total_locked

        # ignore data['unbonding'] - it's still staked

        b_reserved = safe_opt_decimal(data['reserved'])
        b_vesting = (
            to_decimal(data['vesting']['total_locked'])
            if data['vesting']
            else Decimal(0)
        )

        # do we want to include these?
        # b_democracy = safe_opt_decimal(data['democracy_lock'])
        # b_election = safe_opt_decimal(data['election_lock'])

        b_available = b_total - b_total_locked - b_reserved
        b_locked = b_reserved

        if b_available:
            yield BalanceItem.from_api(
                balance_raw=int(b_available),
                coin=self.coin,
                asset_type=AssetType.AVAILABLE,
                raw=data,
            )

        if b_staked:
            yield BalanceItem.from_api(
                balance_raw=int(b_staked),
                coin=self.coin,
                asset_type=AssetType.STAKED,
                raw=data,
            )

        if b_vesting:
            yield BalanceItem.from_api(
                balance_raw=int(b_vesting),
                coin=self.coin,
                asset_type=AssetType.VESTING,
                raw=data['vesting'],
            )

        if b_locked:
            yield BalanceItem.from_api(
                balance_raw=int(b_locked),
                coin=self.coin,
                asset_type=AssetType.LOCKED,
                raw=data,
            )

    def _get_staking_reward(self, address: str) -> Optional[BalanceItem]:
        reward = sum(self._yield_staking_rewards(address))
        if not reward:
            return

        return BalanceItem.from_api(
            balance_raw=int(reward),
            coin=self.coin,
            asset_type=AssetType.REWARDS,
            raw={},
        )

    def _yield_staking_rewards(self, address: str) -> Iterable[Decimal]:
        body = {
            'address': address,
            'row': self.api_options.max_items_per_page,
            'page': 0,
        }
        continuous_count = 0

        while True:
            response = self._post('get_rewards', body=json.dumps(body))
            total_count = response['data']['count']
            continuous_count += len(response['data']['list'])

            for i in response['data']['list']:
                yield (
                    to_decimal(i['amount'])
                    if i['event_id'] == 'Rewarded'
                    else to_decimal(i['amount']) * -1
                )

            if continuous_count >= total_count:
                return

            body['page'] += self.api_options.page_offset_step

    def _post(self, request_method: str, body):
        headers = {'Content-Type': 'application/json'}
        return self.post(request_method, body=body, headers=headers)

    def _opt_raise_on_other_error(self, response: Response) -> None:
        json_response = response.json()
        code = json_response.get('code')

        if code == 10004:
            raise InvalidAddressException('Invalid address format.')
        elif code:
            raise ApiException(json_response['message'])

    @staticmethod
    def _get_api_options(blockchain: Blockchain, base_url: str) -> ApiOptions:
        return ApiOptions(
            blockchain=blockchain,
            base_url=base_url,
            start_offset=0,
            max_items_per_page=100,
            page_offset_step=1,
        )


class PolkadotSubscanApi(SubscanApi):
    coin = COIN_DOT
    api_options = SubscanApi._get_api_options(
        blockchain=Blockchain.POLKADOT, base_url='https://polkadot.api.subscan.io'
    )


class KusamaSubscanApi(SubscanApi):
    coin = COIN_KSM
    api_options = SubscanApi._get_api_options(
        blockchain=Blockchain.KUSAMA, base_url='https://kusama.api.subscan.io'
    )



from abc import ABCMeta
from decimal import Decimal
from typing import Iterable, Optional

from blockapi.utils.num import to_decimal
from blockapi.v2.base import ApiOptions, BlockchainApi, IBalance
from blockapi.v2.coins import COIN_ATOM
from blockapi.v2.models import AssetType, BalanceItem, Blockchain, Coin


class CosmosApiBase(BlockchainApi, IBalance, metaclass=ABCMeta):
    """
    Cosmos RPC
    API docs: https://cosmos.network/rpc/
    Explorer: https://www.mintscan.io
    """

    coin = COIN_ATOM

    API_BASE_URL = 'https://lcd-cosmos.cosmostation.io/'
    API_BASE_RATE_LIMIT = 0.2

    supported_requests = {
        'get_balances': '/cosmos/bank/v1beta1/balances/{address}',
        'get_staked_balance': (
            '/cosmos/staking/v1beta1/delegations/{address}?pagination.limit=1000'
        ),
        'get_unbonding_balance': (
            '/cosmos/staking/v1beta1/delegators/{address}/unbonding_delegations'
        ),
        'get_rewards': 'cosmos/distribution/v1beta1/delegators/{address}/rewards',
        # TODO support commissions, valid only for validators?
        'get_commission': (
            'cosmos/distribution/v1beta1/validators/{validator_address}/commission'
        ),
    }

    _tokens_map: Optional[dict[str, dict]] = None

    @property
    def tokens_map(self) -> dict[str, dict]:
        return NotImplemented

    def get_balance(self, address: str) -> list[BalanceItem]:
        balances = []

        balances.extend(list(self._yield_available_balance(address)))
        if staked_balance := self._get_staked_balance(address):
            balances.append(staked_balance)
        if reward_balance := self._get_reward_balance(address):
            balances.append(reward_balance)

        return balances

    def _yield_available_balance(self, address: str) -> Iterable[BalanceItem]:
        response = self._get('get_balances', address=address)

        for b in response['balances']:
            if b['denom'] == self.coin.address:
                yield BalanceItem.from_api(
                    balance_raw=b['amount'], coin=self.coin, raw=b
                )
            else:
                if b['denom'] in self.tokens_map:
                    token = self._get_token_data(b['denom'])
                else:
                    token = Coin.from_api(
                        blockchain=Blockchain.COSMOS,
                        decimals=self.coin.decimals,
                        address=b['denom'],
                    )
                yield BalanceItem.from_api(balance_raw=b['amount'], coin=token, raw=b)

    def _get_token_data(self, denom: str) -> Coin:
        raw_token = self.tokens_map[denom]
        return Coin(
            symbol=raw_token['dp_denom'],
            name=raw_token['dp_denom'],
            decimals=raw_token['decimal'],
            blockchain=Blockchain.COSMOS,
            address=raw_token['base_denom'],
            standards=raw_token['type'],
        )

    def _get_staked_balance(self, address: str) -> Optional[BalanceItem]:
        staked_response = self._get('get_staked_balance', address=address)
        staked_balance = sum(
            to_decimal(d['balance']['amount'])
            for d in staked_response['delegation_responses']
        )

        unbonding_response = self._get('get_unbonding_balance', address=address)
        unbonding_balance = sum(
            to_decimal(unbonding_entry['balance'])
            for unbonding in unbonding_response['unbonding_responses']
            for unbonding_entry in unbonding['entries']
        )

        # unbonding balance is still staked
        staked_balance += unbonding_balance
        if not staked_balance:
            return

        return BalanceItem.from_api(
            balance_raw=str(staked_balance),
            coin=self.coin,
            asset_type=AssetType.STAKED,
            raw={
                'delegated': staked_response['delegation_responses'],
                'unbonding': unbonding_response['unbonding_responses'],
            },
        )

    def _get_reward_balance(self, address: str) -> Optional[BalanceItem]:
        response = self._get('get_rewards', address=address)
        if not response.get('rewards'):
            return

        return BalanceItem.from_api(
            balance_raw=response['total'][0]['amount'],
            coin=self.coin,
            asset_type=AssetType.REWARDS,
            raw=response['total'],
        )

    def _get_commission(self, validator_address: str) -> Decimal:
        response = self._get('get_commission', validator_address=validator_address)
        return to_decimal(response['commission']['commission'][0]['amount'])

    def _get(self, request_method: str, **req_args) -> dict:
        headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) '
            'AppleWebKit/537.36 (KHTML, like Gecko) '
            'Chrome/50.0.2661.102 Safari/537.36',
            'Referer': 'https://www.mintscan.io/',
        }
        return self.get(request_method, headers, **req_args)


class CosmosApi(CosmosApiBase):
    coin = COIN_ATOM
    api_options = ApiOptions(
        blockchain=Blockchain.COSMOS,
        base_url=CosmosApiBase.API_BASE_URL,
        rate_limit=CosmosApiBase.API_BASE_RATE_LIMIT,
    )

    @property
    def tokens_map(self) -> dict[str, dict]:
        if self._tokens_map is None:
            response = self._session.get('https://api.mintscan.io/v2/assets/cosmos')
            token_list = response.json()
            if 'assets' not in token_list:
                self._tokens_map = {}
            else:
                self._tokens_map = {a['denom']: a for a in token_list['assets']}

        return self._tokens_map



from typing import Dict, List

from requests import Response

from blockapi.v2.base import ApiException, ApiOptions, BlockchainApi, IBalance
from blockapi.v2.coins import COIN_ETH
from blockapi.v2.models import BalanceItem, Blockchain


class OptimismEtherscanApi(BlockchainApi, IBalance):
    """
    Optimism
    Explorer: https://optimistic.etherscan.io
    """

    coin = COIN_ETH
    api_options = ApiOptions(
        blockchain=Blockchain.OPTIMISM,
        base_url='https://api-optimistic.etherscan.io/api',
        rate_limit=0.2,  # 0.1 in case of api_key
    )

    supported_requests = {
        'get_balance': '?module=account&action=balance&address={address}&tag=latest&apikey={api_key}'
    }

    def __init__(self, api_key: str = ''):
        super().__init__(api_key)

    def _parse_eth_balance(self, response: Dict) -> BalanceItem:
        return BalanceItem.from_api(
            balance_raw=response.get('result', 0),
            coin=self.coin,
            raw=response,
        )

    def get_balance(self, address: str) -> List[BalanceItem]:
        # TODO: currently returns only ETH balance (not all ERC20 tokens)
        response = self.get(
            'get_balance',
            address=address,
            api_key=self.api_key,
            # API requires User-Agent.
            headers={
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, '
                'like Gecko) Chrome/50.0.2661.102 Safari/537.36'
            },
        )

        return [self._parse_eth_balance(response)]

    def _opt_raise_on_other_error(self, response: Response) -> None:
        json_response = response.json()
        if json_response["message"] == "OK":
            return

        raise ApiException(json_response['result'])



import logging
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Union

from eth_utils import to_checksum_address
from pydantic import BaseModel, validator

from blockapi.utils.num import decimals_to_raw
from blockapi.v2.api.debank_maps import (
    ALL_COINS,
    DEBANK_ASSET_TYPES,
    NATIVE_COIN_MAP,
    REWARD_ASSET_TYPE_MAP,
)
from blockapi.v2.base import ApiOptions, CustomizableBlockchainApi, IBalance, IPortfolio
from blockapi.v2.blockchain_mapping import get_blockchain_from_debank_chain
from blockapi.v2.models import (
    AssetType,
    BalanceItem,
    Blockchain,
    Coin,
    CoinInfo,
    Pool,
    Protocol,
    TokenRole,
)

logger = logging.getLogger(__name__)


class DebankModelBalanceItem(BaseModel):
    id: str
    chain: str
    name: str
    symbol: str
    display_symbol: Optional[str]
    optimized_symbol: Optional[str]
    decimals: int
    logo_url: Optional[str]
    protocol_id: Optional[str]
    time_at: Optional[float]
    amount: float
    raw_amount: Optional[float]
    raw_value: Optional[dict]


class DebankModelPoolItemDetail(BaseModel):
    description: Optional[str]
    health_rate: Optional[float]
    unlock_at: Optional[float]
    token_list: Optional[List[Dict]]
    supply_token_list: Optional[List[Dict]]
    borrow_token_list: Optional[List[Dict]]
    reward_token_list: Optional[List[Dict]]


class DebankModelPoolItem(BaseModel):
    id: str
    project_id: str
    adapter_id: str


class DebankModelPortfolioItem(BaseModel):
    name: str
    detail: DebankModelPoolItemDetail
    pool_id: Optional[str]
    pool: Optional[DebankModelPoolItem]

    @validator('pool')
    def require_pool_or_pool_id(cls, v, values, **kwargs):
        if v is not None:
            return v

        if not 'pool_id' in values or values['pool_id'] is None:
            raise ValueError('either pool or pool_id must have a value')

        return v


class DebankModelProtocol(BaseModel):
    id: str
    chain: str
    name: str
    site_url: Optional[str]
    logo_url: Optional[str]
    has_supported_portfolio: Optional[bool] = False
    tvl: Optional[float]


class DebankModelPortfolio(DebankModelProtocol):
    portfolio_item_list: List[DebankModelPortfolioItem]


class DebankProtocolParser:
    def parse(self, response: List) -> Dict[str, Protocol]:
        protocols = {}
        for item in response:
            model = DebankModelProtocol(**item)
            protocol = self.parse_item(model)
            if protocol:
                protocols[protocol.protocol_id] = protocol

        return protocols

    @staticmethod
    def parse_item(item: DebankModelProtocol) -> Optional[Protocol]:
        blockchain = get_blockchain_from_debank_chain(item.chain)
        if not blockchain:
            logger.warning(f'No blockchain found for protocol {item.id}')
            return None

        return Protocol.from_api(
            protocol_id=item.id,
            chain=blockchain,
            name=item.name,
            user_deposit=item.tvl,
            site_url=item.site_url,
            logo_url=item.logo_url,
            has_supported_portfolio=item.has_supported_portfolio,
        )


class DebankProtocolCache:
    def __init__(self, timeout: int = 3600):
        self._timeout: int = timeout
        self._data: Dict[str, Protocol] = {}
        self._timelimit = datetime.now()

    def invalidate(self):
        self._timelimit = datetime.now()

    def needs_update(self) -> bool:
        return datetime.now() >= self._timelimit

    def update(self, data: Dict[str, Protocol]) -> None:
        self._timelimit = datetime.now() + timedelta(seconds=self._timeout)
        self._data = data

    def get(self, key: str) -> Optional[Protocol]:
        protocol = self._data.get(key)
        if protocol is None:
            logger.debug("Protocol '%s' not found.", key)

        return protocol


class DebankBalanceParser:
    def __init__(self, protocol_cache: DebankProtocolCache):
        self._protocols = protocol_cache

    def parse(
        self,
        response: List,
        asset_type: AssetType = AssetType.AVAILABLE,
        is_wallet: bool = True,
        token_set: Optional[List[str]] = None,
        token_role: Optional[TokenRole] = None,
        pool_id: Optional[str] = None,
    ) -> List[BalanceItem]:
        items = []
        for item in response:
            balance_item = DebankModelBalanceItem(**item)
            balance_item.raw_value = item

            balance = self.parse_item(
                balance_item, asset_type, is_wallet, token_set, token_role, pool_id
            )
            if balance is not None:
                items.append(balance)

        return items

    def parse_item(
        self,
        balance_item: DebankModelBalanceItem,
        asset_type: AssetType = AssetType.AVAILABLE,
        is_wallet: bool = True,
        token_set: Optional[List[str]] = None,
        token_role: Optional[TokenRole] = None,
        pool_id: Optional[str] = None,
    ) -> Optional[BalanceItem]:
        raw_amount = balance_item.raw_amount or 0
        amount = balance_item.amount or 0

        if raw_amount == 0 and amount == 0:
            logger.debug(
                "Skipping balance item: '%s' - balance is zero.",
                balance_item.name,
            )

            return None

        symbol = self.get_symbol(balance_item)
        coin = self.get_coin(balance_item, symbol)

        if not coin:
            logger.error(
                f'DeBank: Skipping balance - could not parse coin "{balance_item.id} {balance_item.chain} {balance_item.symbol}". Amount={amount}'
            )
            return None

        if not coin.blockchain:
            logger.error(
                f'DeBank: Skipping balance - could not parse blockchain "{balance_item.chain}". Amount={amount} (raw={raw_amount})'
            )
            return None

        if asset_type == AssetType.INVESTMENT and amount < 0:
            asset_type = AssetType.DEBT
            amount = -amount

        if raw_amount == 0 and amount != 0:
            raw_amount = decimals_to_raw(amount, coin.decimals)

        balance = BalanceItem.from_api(
            balance_raw=raw_amount,
            coin=coin,
            asset_type=asset_type,
            last_updated=balance_item.time_at,
            raw=balance_item.raw_value,
            protocol=self._protocols.get(balance_item.protocol_id),
            is_wallet=is_wallet,
            token_set=token_set,
            token_role=token_role,
            pool_id=pool_id,
        )

        return balance

    def get_coin(
        self, balance_item: DebankModelBalanceItem, symbol: str
    ) -> Optional[Coin]:
        address = balance_item.id
        blockchain = get_blockchain_from_debank_chain(balance_item.chain)
        coin = NATIVE_COIN_MAP.get(address)

        if coin is not None and coin.blockchain == blockchain and coin.symbol == symbol:
            return coin

        valid_address = make_checksum_address(address)

        if valid_address:
            return Coin.from_api(
                symbol=symbol,
                name=balance_item.name,
                decimals=balance_item.decimals,
                blockchain=blockchain,
                address=valid_address,
                standards=[],
                protocol_id=balance_item.protocol_id,
                info=CoinInfo(logo_url=balance_item.logo_url),
            )

        base_coin = ALL_COINS.get(symbol.lower())
        if base_coin:
            return Coin.from_api(
                symbol=base_coin.symbol,
                name=base_coin.name,
                decimals=balance_item.decimals,
                blockchain=blockchain,
                address=base_coin.address,
                standards=base_coin.standards,
                protocol_id=balance_item.protocol_id,
                info=base_coin.info,
            )

        return None

    @staticmethod
    def get_symbol(raw_balance: DebankModelBalanceItem) -> str:
        return (
            raw_balance.display_symbol
            or raw_balance.optimized_symbol
            or raw_balance.symbol
        )


def make_checksum_address(address: str) -> Optional[str]:
    try:
        return to_checksum_address(address)
    except ValueError as e:
        logger.warning(f'Cannot parse address "{address}", error: {e}')
        return None


class DebankPortfolioParser:
    def __init__(
        self, protocol_parser: DebankProtocolParser, balance_parser: DebankBalanceParser
    ):
        self._protocol_parser = protocol_parser
        self._balance_parser = balance_parser

    def parse(self, response: List) -> List[Pool]:
        items = []
        for item in response:
            portfolio = DebankModelPortfolio(**item)
            parsed = self.parse_items(portfolio)
            items.extend(parsed)

        return items

    def parse_items(self, raw_portfolio: DebankModelPortfolio) -> List[Pool]:
        root_protocol = self._protocol_parser.parse_item(raw_portfolio)
        pools = self._parse_portfolio_item_list(
            raw_portfolio.portfolio_item_list or [], root_protocol
        )

        return pools

    def _parse_portfolio_item_list(
        self,
        raw_portfolio_items: List[DebankModelPortfolioItem],
        root_protocol: Protocol,
    ) -> List[Pool]:
        items = []
        pools = {}
        for item in raw_portfolio_items:
            pool = self._parse_portfolio_item(item, root_protocol, pools)
            if pool.pool_id:
                pools[pool.pool_id] = pool

            items.append(pool)

        return items

    def _parse_portfolio_item(
        self,
        item: DebankModelPortfolioItem,
        pool_protocol: Protocol,
        pools: Dict[str, Pool],
    ) -> Pool:
        pool_id = self._get_pool_id(item)
        detail = item.detail

        health_rate = detail.health_rate
        locked_until = detail.unlock_at

        asset_type = self._parse_asset_type(item.name)
        borrow_type = self._get_borrow_asset_type(asset_type)
        reward_type = self._get_reward_asset_type(asset_type)

        items = []

        tokens = self._update_items(
            items,
            detail.supply_token_list,
            asset_type,
            token_role=TokenRole.SUPPLY,
            pool_id=pool_id,
        )
        self._update_items(
            items,
            detail.borrow_token_list,
            borrow_type,
            tokens,
            token_role=TokenRole.BORROW,
            pool_id=pool_id,
        )
        self._update_items(
            items,
            detail.reward_token_list,
            reward_type,
            tokens,
            token_role=TokenRole.REWARD,
            pool_id=pool_id,
        )
        self._update_items(items, detail.token_list, asset_type, pool_id=pool_id)

        pool = pools.get(pool_id)

        if pool is None:
            pool = Pool.from_api(
                pool_id=pool_id,
                protocol=pool_protocol,
                locked_until=locked_until,
                health_rate=health_rate,
                items=items,
                token_set=detail.description,
                project_id=item.pool.project_id if item.pool else None,
                adapter_id=item.pool.adapter_id if item.pool else None,
            )
        else:
            pool.append_items(items)

        return pool

    def _update_items(
        self,
        items: List[BalanceItem],
        raw_balances: list[dict],
        asset_type: AssetType,
        tokens: Optional[List[str]] = None,
        token_role: Optional[TokenRole] = None,
        pool_id: Optional[str] = None,
    ):
        if not raw_balances:
            return

        if tokens is None:
            tokens = sorted(
                list(
                    set(
                        [
                            self._balance_parser.get_symbol(DebankModelBalanceItem(**b))
                            for b in raw_balances
                        ]
                    )
                )
            )

        balances = self._balance_parser.parse(
            raw_balances,
            asset_type,
            False,
            token_set=tokens if len(raw_balances) > 1 else None,
            token_role=token_role,
            pool_id=pool_id,
        )

        items.extend(balances)
        return tokens

    @staticmethod
    def _get_pool_id(item: DebankModelPortfolioItem) -> str:
        return item.pool.id if item.pool else item.pool_id

    @staticmethod
    def _parse_asset_type(type_: str) -> Optional[AssetType]:
        # list of valid types: https://docs.open.debank.com/en/reference/api-models/portfolioitemobject
        if type_ is None:
            return AssetType.AVAILABLE

        try:
            lower = type_.lower()
            asset_type = DEBANK_ASSET_TYPES.get(lower)
            if asset_type:
                return asset_type

            return AssetType(lower)
        except ValueError as ve:
            logger.error(ve)
            return AssetType.AVAILABLE

    @staticmethod
    def _get_borrow_asset_type(asset_type):
        if asset_type == AssetType.LENDING:
            return AssetType.LENDING_BORROW

        return asset_type

    @staticmethod
    def _get_reward_asset_type(asset_type):
        return REWARD_ASSET_TYPE_MAP.get(asset_type, asset_type)


class DebankApi(CustomizableBlockchainApi, IBalance, IPortfolio):
    """
    DeBank OpenApi: https://open.debank.com/
    """

    API_BASE_URL = 'https://pro-openapi.debank.com'
    API_BASE_RATE_LIMIT = 0.05  # 20 req / s

    api_options = ApiOptions(
        blockchain=Blockchain.ETHEREUM,
        base_url=API_BASE_URL,
        rate_limit=API_BASE_RATE_LIMIT,
    )

    coin = None

    supported_requests = {
        'get_balance': '/v1/user/all_token_list?id={address}&is_all={is_all}',
        'get_portfolio': '/v1/user/all_complex_protocol_list?id={address}',
        'get_protocols': '/v1/protocol/all_list',
    }

    default_protocol_cache = DebankProtocolCache()

    def __init__(
        self,
        api_key: str,
        is_all: bool,
        protocol_cache: Optional[DebankProtocolCache] = None,
        base_url: Optional[str] = None,
    ):
        super().__init__(base_url)

        self._is_all = bool(is_all)
        self._headers = {'AccessKey': api_key}
        self._protocol_cache = protocol_cache or self.default_protocol_cache
        self._balance_parser = DebankBalanceParser(self._protocol_cache)
        self._protocol_parser = DebankProtocolParser()
        self._portfolio_parser = DebankPortfolioParser(
            self._protocol_parser, self._balance_parser
        )

    def get_balance(self, address: str) -> List[BalanceItem]:
        self._maybe_update_protocols()
        response = self.get(
            'get_balance', headers=self._headers, address=address, is_all=self._is_all
        )
        if self._has_error(response):
            return []

        return self._balance_parser.parse(response)

    def get_protocols(self) -> Dict[str, Protocol]:
        response = self.get('get_protocols', headers=self._headers)
        if self._has_error(response):
            return {}

        return self._protocol_parser.parse(response)

    def get_portfolio(self, address: str) -> List[Pool]:
        self._maybe_update_protocols()
        response = self.get('get_portfolio', headers=self._headers, address=address)
        if self._has_error(response):
            return []

        return self._portfolio_parser.parse(response)

    def _maybe_update_protocols(self):
        if self._protocol_cache.needs_update():
            self._protocol_cache.update(self.get_protocols())

    @staticmethod
    def _has_error(response: Union[List, Dict]) -> bool:
        if isinstance(response, list):
            return False

        error = response.get('errors')
        message = response.get('message')
        if message is not None:
            logger.error('DebankApi Error: %s', message)

        if error is not None:
            err_id = error.get('id')
            if err_id is not None:
                logger.error(err_id)

        return message is not None or error is not None

    def __repr__(self):
        return f"{self.__class__.__name__}"



import logging
from abc import ABC
from decimal import Decimal
from functools import lru_cache
from typing import Dict, Iterable, List, Tuple

import requests
from bs4 import BeautifulSoup
from eth_typing import ChecksumAddress
from marko import Markdown
from marko.block import Heading, HTMLBlock
from typing_extensions import TypedDict
from web3 import Web3

from blockapi.utils.num import raw_to_decimals, safe_decimal, to_decimal
from blockapi.v2.api.synthetix.synthetix_abi import (
    erc20_abi,
    exchangerates_abi,
    feepool_abi,
    rewards_escrow_v2_abi,
    staking_rewards_abi,
    synthetix_abi,
    system_settings_abi,
)
from blockapi.v2.api.web3_utils import (
    easy_call,
    ensure_checksum_address,
    get_eth_client,
)
from blockapi.v2.base import BlockchainApi, IBalance
from blockapi.v2.coins import COIN_SNX
from blockapi.v2.models import AssetType, BalanceItem, Blockchain, Coin

logger = logging.getLogger(__name__)


class CollateralizationStats(TypedDict):
    collateralization_ratio: Decimal
    collateralization_ratio_perc: Decimal


class StakingToken(TypedDict):
    symbol: str
    staked: Decimal
    rewards: Decimal


class WeeklyReward(TypedDict):
    exchange: Decimal
    staking: Decimal


class Rewards(TypedDict):
    prev_week: WeeklyReward
    this_week: WeeklyReward


class Staking(CollateralizationStats):
    transferable: Decimal
    debt: Decimal
    staked: Decimal
    vesting: Decimal
    collateral: Decimal
    rewards: Rewards
    staking_tokens: List[StakingToken]


class Synth(TypedDict):
    symbol: str
    contract_address: str


# noinspection PyBroadException
@lru_cache(maxsize=128)
def snx_contract_address(
    contract_name: str, network: str = "mainnet"
) -> ChecksumAddress:
    """
    Dynamically converts Synthetix contract name into its
    Ethereum address.
    """
    if network == 'optimism':
        return snx_optimism_contract_address(contract_name)

    base = "https://contracts.synthetix.io"
    url = (
        f"{base}/{contract_name}"
        if network == "mainnet"
        else f"{base}/{network}/{contract_name}"
    )

    result = requests.get(url)

    # contract address is obtained from redirected etherscan url,
    # as last 42 characters
    try:
        contract_address = result.url[-42:]
        return Web3.toChecksumAddress(contract_address)
    except Exception:
        raise ValueError(f'Contract {contract_name} not found.')


# noinspection PyBroadException
def snx_optimism_contract_address(
    contract_name: str,
) -> ChecksumAddress:
    """
    Dynamically converts Synthetix contract name into its
    Ethereum address on Optimism L2.
    """
    try:
        response = requests.get(
            'https://raw.githubusercontent.com/Synthetixio/synthetix-docs/master/content/addresses.md'
        )
        md = Markdown().parse(response.text)

        # find table with Optimism contracts
        html_tab_raw = None
        return_table = False
        for child in md.children:
            if return_table and type(child) is HTMLBlock:
                html_tab_raw = child.children
                break
            elif (
                type(child) is Heading
                and child.children[0].children == 'MAINNET Optimism (L2)'
            ):
                return_table = True

        table = BeautifulSoup(html_tab_raw, 'lxml')
        row = table.find('td', text=contract_name).parent
        return Web3.toChecksumAddress(row.contents[5].text.strip())

    except Exception:
        raise ValueError(f'Contract {contract_name} not found.')


def get_staking_tokens(network: str) -> List[Tuple[str, str]]:
    """
    Get supported staking tokens for network.
    TODO add testnet networks.
    """
    # noinspection SpellCheckingInspection
    return (
        [
            ('iBTC', 'StakingRewardsiBTC'),
            ('iETH', 'StakingRewardsiETH'),
            ('sTSLA', 'StakingRewardssTSLABalancer'),
        ]
        if network == 'mainnet'
        else []
    )


class SynthetixApi(BlockchainApi, IBalance, ABC):
    decimals: Decimal = Decimal('18')
    coin = COIN_SNX

    def __init__(self, network: str, api_url: str):
        super().__init__()
        self.network = network
        self.blockchain = self.get_blockchain(self.network)
        self.w3 = get_eth_client(api_url)

    def get_balance(self, address: str) -> List[BalanceItem]:
        address = ensure_checksum_address(address)
        return list(self.yield_balances(address))

    def yield_balances(self, address) -> Iterable[BalanceItem]:
        """
        Fetch all balances for snx and synth tokens.
        """
        logger.info("Called Synthetix.yield_balances for address: %s", address)

        staking = self.fetch_staking(address)

        if staking['transferable']:
            yield self._create_balance(
                'SNX', AssetType.AVAILABLE, staking['transferable']
            )

        if staking['debt']:
            yield self._create_balance('sUSD', AssetType.DEBT, staking['debt'])

        if staking['staked']:
            yield self._create_balance('SNX', AssetType.STAKED, staking['staked'])

        if staking['vesting']:
            yield self._create_balance('SNX', AssetType.VESTING, staking['vesting'])

        rewards = Decimal("0")
        for token in staking.get('staking_tokens'):
            if token['staked']:
                yield self._create_balance(
                    token['symbol'], AssetType.STAKED, token['staked']
                )

            if token["rewards"]:
                rewards += token["rewards"]

        if rewards:
            yield self._create_balance('SNX', AssetType.CLAIMABLE, rewards)

    def _create_balance(
        self, symbol: str, asset_type: AssetType, balance_raw: Decimal
    ) -> BalanceItem:
        """
        Create balance item.
        """

        coin = self._get_coin(symbol)

        return BalanceItem(
            balance_raw=balance_raw,
            balance=balance_raw * pow(10, -self.decimals),
            coin=coin,
            asset_type=asset_type,
            raw={},
        )

    def _get_coin(self, symbol: str) -> Coin:
        return Coin(
            symbol=symbol,
            blockchain=self.blockchain,
            name=symbol,
            decimals=int(self.decimals),
            address=self._get_synth_contract(symbol),
        )

    @staticmethod
    def get_blockchain(network: str) -> Blockchain:
        return Blockchain.OPTIMISM if network == 'optimism' else Blockchain.ETHEREUM

    # noinspection PyTypeChecker
    def fetch_staking(self, address: str) -> Staking:
        """
        Fetch Synthetix staking info.
        """
        staking_tokens = get_staking_tokens(self.network)

        ratio = self.get_collateralization_ratio(address)
        return {
            **ratio,
            'transferable': self.get_snx_transferable_amount(address),
            'debt': self.get_total_debt_owed(address),
            'staked': self.compute_snx_staked_amount(address),
            'vesting': self.get_total_escrowed_amount(address),
            'collateral': self.get_collateral(address),
            'rewards': self.get_fees_and_rewards(address),
            'staking_tokens': list(
                self.yield_token_staking_data(address, staking_tokens)
            ),
        }

    def get_snx_transferable_amount(self, address: str) -> Decimal:
        """
        Returns total amount of unlocked SNX tokens.
        """
        snx_contract = self.w3.eth.contract(
            snx_contract_address('Synthetix', self.network), abi=synthetix_abi
        )
        amount = easy_call(snx_contract, 'transferableSynthetix', address)
        return to_decimal(amount)

    def get_total_debt_owed(self, address: str) -> Decimal:
        """
        Returns total debt owed in sUSD.
        """
        snx_contract = self.w3.eth.contract(
            snx_contract_address('Synthetix', self.network), abi=synthetix_abi
        )
        debt = easy_call(snx_contract, 'debtBalanceOf', address, b'sUSD')
        return to_decimal(debt)

    @lru_cache(maxsize=128)
    def get_collateralization_ratio(self, address) -> CollateralizationStats:
        """
        Returns collateralization ratio in %.
        """
        snx_contract = self.w3.eth.contract(
            snx_contract_address('Synthetix', self.network), abi=synthetix_abi
        )
        c_ratio = easy_call(snx_contract, 'collateralisationRatio', address)
        if c_ratio:
            ratio = raw_to_decimals(c_ratio, int(self.decimals))
            return {
                'collateralization_ratio': ratio,
                'collateralization_ratio_perc': 1 / ratio * 100,
            }
        else:
            return {
                'collateralization_ratio': Decimal("0"),
                'collateralization_ratio_perc': Decimal("0"),
            }

    @lru_cache(maxsize=512)
    def get_collateral(self, address: str) -> Decimal:
        """
        Return collateral.
        """
        snx_contract = self.w3.eth.contract(
            snx_contract_address('Synthetix', self.network), abi=synthetix_abi
        )
        collateral = easy_call(snx_contract, 'collateral', address)
        return to_decimal(collateral)

    def compute_snx_staked_amount(self, address: str) -> Decimal:
        """
        Return total staked snx amount.
        """
        system_contract = self.w3.eth.contract(
            snx_contract_address('SystemSettings', self.network),
            abi=system_settings_abi,
        )
        i_ratio = easy_call(system_contract, 'issuanceRatio')
        i_ratio = raw_to_decimals(i_ratio, int(self.decimals))

        collateral = self.get_collateral(address)
        collateralization_ratio = self.get_collateralization_ratio(address)[
            "collateralization_ratio"
        ]

        return collateral * min(Decimal("1"), collateralization_ratio / i_ratio)

    def get_fees_and_rewards(self, address: str) -> Rewards:
        """
        Returns the available Synth exchange rewards (fees, in sUSD) and
        SNX staking rewards (in SNX) as 3 tuples, current one (not yet
        claimable), one week ago, two weeks ago
        These values (except current ones) are zeroed as soon as
        the rewards are claimed!
        The current one is zeroed at the beginning of next fee period.
        """
        fee_pool_contract = self.w3.eth.contract(
            snx_contract_address('FeePool', self.network), abi=feepool_abi
        )
        fees = easy_call(fee_pool_contract, 'feesByPeriod', address)

        return {
            "prev_week": {
                "exchange": to_decimal(fees[1][0]),
                "staking": to_decimal(fees[1][1]),
            },
            "this_week": {
                "exchange": to_decimal(fees[0][0]),
                "staking": to_decimal(fees[0][1]),
            },
        }

    def yield_token_staking_data(
        self, address: str, synths: List[Tuple[str, str]]
    ) -> Iterable[StakingToken]:
        for symbol, contract_name in synths:
            staking_contract = self.w3.eth.contract(
                snx_contract_address(contract_name, self.network),
                abi=staking_rewards_abi,
            )

            staked = easy_call(staking_contract, 'balanceOf', address)
            rewards = easy_call(staking_contract, 'earned', address)
            yield {
                'symbol': symbol,
                'staked': to_decimal(staked),
                'rewards': to_decimal(rewards),
            }

    @lru_cache()
    def _get_synth_contract(self, symbol: str) -> str:
        """
        Get synth contract address.
        """
        snx_contract = self.w3.eth.contract(
            snx_contract_address('Synthetix', self.network), abi=synthetix_abi
        )

        if symbol == 'SNX':
            synth_addr = easy_call(snx_contract, 'proxy')
        else:
            synth_contract_addr = Web3.toChecksumAddress(
                easy_call(snx_contract, 'synths', symbol.encode())
            )
            synth_contract = self.w3.eth.contract(synth_contract_addr, abi=erc20_abi)
            synth_addr = easy_call(synth_contract, 'proxy')

        return synth_addr

    def get_total_escrowed_amount(self, address: str) -> Decimal:
        """
        Return number of SNX tokens escrowed.
        """
        rewards_escrow_contract = self.w3.eth.contract(
            snx_contract_address('RewardEscrowV2', self.network),
            abi=rewards_escrow_v2_abi,
        )
        total_escrowed = easy_call(
            rewards_escrow_contract,
            "totalEscrowedAccountBalance",
            address,
        )

        return to_decimal(total_escrowed)

    def get_token_xchg_rates(self, synths: List[Synth]) -> Dict:
        """
        Reads Synthetix tokens exchange rates needed for further
        calculations.
        """
        xchg_contract = self.w3.eth.contract(
            snx_contract_address('ExchangeRates', self.network),
            abi=exchangerates_abi,
        )
        xchg_rates = {}

        for synth in synths:
            symbol = (
                synth['symbol'][0].lower() + synth['symbol'][1:4]
                if len(synth['symbol']) >= 4
                else synth['symbol']
            )

            xchg_rates[synth['contract_address']] = raw_to_decimals(
                easy_call(
                    xchg_contract,
                    'rateForCurrency',
                    symbol.encode('utf-8'),
                ),
                decimals=int(self.decimals),
            )

        xchg_rates['XDR'] = safe_decimal(
            easy_call(xchg_contract, 'effectiveValue', b'XDR', 1, b'sUSD')
        )

        return xchg_rates


class SynthetixMainnetApi(SynthetixApi):
    def __init__(
        self,
        api_url: str,
    ):
        super().__init__(network="mainnet", api_url=api_url)


class SynthetixOptimismApi(SynthetixApi):
    def __init__(
        self,
        api_url: str,
    ):
        super().__init__(network='optimism', api_url=api_url)



from .models import Blockchain, Coin, CoinInfo

COIN_ETH = Coin(
    symbol='ETH',
    name='Ethereum',
    decimals=18,
    blockchain=Blockchain.ETHEREUM,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(coingecko_id='ethereum'),
)


COIN_SOL = Coin(
    symbol='SOL',
    name='Solana',
    decimals=9,
    blockchain=Blockchain.SOLANA,
    info=CoinInfo(
        logo_url='https://raw.githubusercontent.com/solana-labs/token-list/main/assets/'
        'mainnet/So11111111111111111111111111111111111111112/logo.png',
        coingecko_id='solana',
        website='https://solana.com/',
    ),
)


COIN_TERRA = Coin(
    symbol='LUNA',
    name='Terra',
    decimals=6,
    blockchain=Blockchain.TERRA,
    address='uluna',
    standards=['terra-native'],
    info=CoinInfo(
        logo_url='https://assets.terra.money/icon/60/Luna.png',
        coingecko_id='terra-luna',
        website='https://www.terra.money/',
    ),
)


COIN_MATIC = Coin(
    symbol='MATIC',
    name='Matic',
    decimals=18,
    blockchain=Blockchain.POLYGON,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='matic-network',
        website='https://polygon.technology/',
    ),
)

COIN_SDN = Coin(
    symbol='SDN',
    name='Shiden Network',
    decimals=18,
    blockchain=Blockchain.ASTAR,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='shiden-network',
        website='https://shiden.astar.network/',
    ),
)

COIN_AVAX = Coin(
    symbol='AVAX',
    name='AVAX',
    decimals=18,
    blockchain=Blockchain.AVALANCHE,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='avalanche-2',
        website='https://www.avax.network/',
    ),
)

COIN_RON = Coin(
    symbol='RON',
    name='Ronin',
    decimals=18,
    blockchain=Blockchain.AXIE,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='ronin',
        website='https://bridge.roninchain.com',
    ),
)

COIN_BNB = Coin(
    symbol='BNB',
    name='Binance Coin',
    decimals=18,
    blockchain=Blockchain.BINANCE_SMART_CHAIN,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='binancecoin',
        website='https://www.binance.com',
    ),
)

COIN_FTM = Coin(
    symbol='FTM',
    name='Fantom',
    decimals=18,
    blockchain=Blockchain.FANTOM,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='fantom',
        website='https://fantom.foundation',
    ),
)

COIN_HT = Coin(
    symbol='HT',
    name='Huobi Token',
    decimals=18,
    blockchain=Blockchain.HECO,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='huobi-token',
        website='https://www.huobi.com',
    ),
)

COIN_IOTX = Coin(
    symbol='IOTX',
    name='IoTeX',
    decimals=18,
    blockchain=Blockchain.IOTEX,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='iotex',
        website='https://iotex.io/',
    ),
)

COIN_KLAY = Coin(
    symbol='KLAY',
    name='Klaytn',
    decimals=18,
    blockchain=Blockchain.KLAY_TOKEN,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(
        coingecko_id='klaytn',
        website='https://www.klaytn.com/',
    ),
)

COIN_MOVR = Coin(
    symbol='MOVR',
    name='Moonriver',
    decimals=18,
    blockchain=Blockchain.MOONBEAM_MOONRIVER,
    address='0x98878b06940ae243284ca214f92bb71a2b032b8a',
    info=CoinInfo(
        coingecko_id='moonriver',
        website='https://moonbeam.network/networks/moonriver/',
    ),
)

COIN_PALM = Coin(
    symbol='PALM',
    name='Palm',
    decimals=18,
    blockchain=Blockchain.PALM,
    address='0x0000000000000000000000000000000000000000',
)

COIN_PERP = Coin(
    symbol='PERP',
    name='Perpetual',
    decimals=0,
    blockchain=Blockchain.ETHEREUM,
    address='0xbC396689893D065F41bc2C6EcbeE5e0085233447',
    info=CoinInfo(
        coingecko_id='perpetual-protocol',
        website='https://perpetual.io/',
    ),
)


COIN_RSK = Coin(
    symbol='RBTC',
    name='Rootstock RSK',
    decimals=18,  # TODO verify this.
    blockchain=Blockchain.ROOTSTOCK,
    address='0x0000000000000000000000000000000000000000',
)

COIN_DOT = Coin(
    symbol='DOT',
    name='Polkadot',
    decimals=10,
    blockchain=Blockchain.POLKADOT,
    info=CoinInfo(coingecko_id='polkadot', tags=['native']),
)

COIN_KSM = Coin(
    symbol='KSM',
    name='Kusama',
    decimals=12,
    blockchain=Blockchain.KUSAMA,
    info=CoinInfo(coingecko_id='kusama', tags=['native']),
)

COIN_ATOM = Coin(
    symbol='ATOM',
    name='Cosmos Hub',
    decimals=6,
    blockchain=Blockchain.COSMOS,
    address='uatom',
    standards=['staking'],
    info=CoinInfo(coingecko_id='cosmos'),
)

COIN_XDAI = Coin(
    symbol='XDAI',
    name='xDai',
    decimals=18,
    blockchain=Blockchain.XDAI,
    info=CoinInfo(coingecko_id='xdai'),
)

COIN_OKT = Coin(
    symbol='OKT',
    name='OKC',
    decimals=18,
    blockchain=Blockchain.OKT,
    info=CoinInfo(coingecko_id='oec-token'),
)

COIN_OP = Coin(
    symbol='OP',
    name='OP',
    decimals=18,
    blockchain=Blockchain.OPTIMISM,
    address='0x4200000000000000000000000000000000000042',
    info=CoinInfo(coingecko_id='optimism'),
)

COIN_CELO = Coin(
    symbol='CELO',
    name='Celo',
    decimals=18,
    blockchain=Blockchain.CELO,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(coingecko_id='celo'),
)

COIN_CRO = Coin(
    symbol='CRO',
    name='Cronos',
    decimals=18,
    blockchain=Blockchain.CRONOS,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(coingecko_id='crypto-com-chain'),
)

COIN_BOBA = Coin(
    symbol='BOBA',
    name='Boba Network',
    decimals=18,
    blockchain=Blockchain.BOBA,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(coingecko_id='boba-network'),
)

COIN_METIS = Coin(
    symbol='METIS',
    name='Metis',
    decimals=18,
    blockchain=Blockchain.METIS_ANDROMEDA,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(coingecko_id='metis-token'),
)

COIN_BTT = Coin(
    symbol='BTT',
    name='BitTorrent',
    decimals=18,
    blockchain=Blockchain.BIT_TORRENT,
    address='TAFjULxiVgT4qWk6UZwjqwZXTSaGaqnVp4',
    info=CoinInfo(coingecko_id='bittorrent'),
)

COIN_AURORA = Coin(
    symbol='AURORA',
    name='Aurora',
    decimals=18,
    blockchain=Blockchain.AURORA,
    address='0xaaaaaa20d9e0e2461697782ef11675f668207961',
    info=CoinInfo(coingecko_id='aurora-near'),
)

COIN_GLMR = Coin(
    symbol='GLMR',
    name='Moonbeam',
    decimals=18,
    blockchain=Blockchain.MOONBEAM,
    info=CoinInfo(coingecko_id='moonbeam'),
)

COIN_FUSE = Coin(
    symbol='FUSE',
    name='Fuse',
    decimals=18,
    blockchain=Blockchain.FUSE,
    address='0x0000000000000000000000000000000000000000',
    info=CoinInfo(coingecko_id='fuse-network-token'),
)

COIN_ONE = Coin(
    symbol='ONE',
    name='Harmony',
    decimals=18,
    blockchain=Blockchain.HARMONY,
    info=CoinInfo(coingecko_id='harmony'),
)

COIN_ASTR = Coin(
    symbol='ASTR',
    name='Astar',
    decimals=18,
    blockchain=Blockchain.ASTAR,
    info=CoinInfo(coingecko_id='astar'),
)

COIN_WAN = Coin(
    symbol='WAN',
    name='Wanchain',
    decimals=18,
    blockchain=Blockchain.WANCHAIN,
    info=CoinInfo(coingecko_id='wanchain'),
)

COIN_KCS = Coin(
    symbol='KCS',
    name='KuCoin',
    decimals=18,
    blockchain=Blockchain.KUCOIN,
    info=CoinInfo(coingecko_id='kucoin-shares'),
)

COIN_SGB = Coin(
    symbol='SGB',
    name='Songbird',
    decimals=18,
    blockchain=Blockchain.SONGBIRD,
    info=CoinInfo(coingecko_id='songbird'),
)

COIN_EVMOS = Coin(
    symbol='EVMOS',
    name='EvmOS',
    decimals=18,
    blockchain=Blockchain.COSMOS,
    info=CoinInfo(coingecko_id='evmos'),
)

COIN_TLOS = Coin(
    symbol='TLOS',
    name='Telos',
    decimals=18,
    blockchain=Blockchain.TELOS,
)

COIN_BOS = Coin(symbol='BOS', name="BosCoin", decimals=7, blockchain=Blockchain.BOS)

COIN_BTC = Coin(
    symbol='BTC',
    name='Bitcoin',
    decimals=8,
    blockchain=Blockchain.BITCOIN,
    info=CoinInfo(coingecko_id='bitcoin'),
)

COIN_LTC = Coin(
    symbol='LTC',
    name='Litecoin',
    decimals=8,
    blockchain=Blockchain.LITECOIN,
    info=CoinInfo(coingecko_id='litecoin'),
)


COIN_DOGE = Coin(
    symbol='DOGE',
    name='Dogecoin',
    decimals=8,
    blockchain=Blockchain.DOGECHAIN,
    info=CoinInfo(coingecko_id='dogecoin'),
)

COIN_SNX = Coin(
    symbol='SNX',
    name='Synthetix',
    decimals=18,
    address='0xc011a73ee8576fb46f5e1c5751ca3b9fe0af2a6f',
    blockchain=Blockchain.ETHEREUM,
    info=CoinInfo(coingecko_id='havven'),
)



from decimal import Decimal

import pytest

from blockapi.test.v2.api.conftest import read_file
from blockapi.v2.api.synthetix import (
    SynthetixMainnetApi,
    SynthetixOptimismApi,
    snx_contract_address,
)
from blockapi.v2.models import AssetType, BalanceItem, Blockchain, Coin

test_address = '0xE2e4F2A725E42D0F0EF6291F46c430F963482001'
contract_name = 'Synthetix'


def mock_redirect(requests_mock, source, target, text):
    requests_mock.get(source, status_code=302, headers={'location': target})
    requests_mock.get(target, text=text)


def test_snx_contract_address_mainnet(requests_mock):
    mock_redirect(
        requests_mock,
        source='https://contracts.synthetix.io/Synthetix',
        target='https://etherscan.io/address/0x08F30Ecf2C15A783083ab9D5b9211c22388d0564',
        text='',
    )

    assert (
        snx_contract_address(contract_name)
        == '0x08F30Ecf2C15A783083ab9D5b9211c22388d0564'
    )


def test_snx_contract_address_optimism(requests_mock):
    text = read_file('synthetix/data/contracts.md')
    requests_mock.get(
        'https://raw.githubusercontent.com/Synthetixio/synthetix-docs/master/content/addresses.md',
        text=text,
    )

    assert (
        snx_contract_address(contract_name, 'optimism')
        == '0x49B35BE7D96888C02F342552aB218d859599aCeb'
    )


def test_invalid_contract_raises(requests_mock):
    requests_mock.get('https://contracts.synthetix.io/abc', status_code=404)

    with pytest.raises(ValueError, match='Contract abc not found'):
        snx_contract_address('abc')


def test_invalid_contract_optimism_raises(requests_mock):
    text = read_file('synthetix/data/contracts.md')
    requests_mock.get(
        'https://raw.githubusercontent.com/Synthetixio/synthetix-docs/master/content/addresses.md',
        text=text,
    )

    with pytest.raises(ValueError, match='Contract abc not found'):
        snx_contract_address('abc', 'optimism')


def test_create_with_custom_api():
    api = SynthetixMainnetApi(api_url='http://localhost:1234/')
    assert api.w3.provider.endpoint_uri == 'http://localhost:1234/'


def test_create_with_custom_optimism_api():
    api = SynthetixOptimismApi(api_url='http://localhost:1234/')
    assert api.w3.provider.endpoint_uri == 'http://localhost:1234/'


@pytest.mark.skip("For some reason VCR cassette doesn't work properly")
@pytest.mark.vcr()
def test_synthetix_optimism_api():
    api = SynthetixOptimismApi(api_url='https://mainnet.optimism.io/')
    balances = api.get_balance(test_address)

    assert balances == [
        BalanceItem(
            balance=Decimal('376576.311381339269627458'),
            balance_raw=Decimal('376576311381339269627458'),
            raw={},
            coin=Coin(
                symbol='sUSD',
                name='sUSD',
                decimals=18,
                blockchain=Blockchain.OPTIMISM,
                address='0x8c6f28f2F1A3C87F0f938b96d27520d9751ec8d9',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.DEBT,
        ),
        BalanceItem(
            balance=Decimal('528051.112125559541484602'),
            balance_raw=Decimal('528051112125559541484602'),
            raw={},
            coin=Coin(
                symbol='SNX',
                name='SNX',
                decimals=18,
                blockchain=Blockchain.OPTIMISM,
                address='0x8700dAec35aF8Ff88c16BdF0418774CB3D7599B4',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.STAKED,
        ),
        BalanceItem(
            balance=Decimal('188153.171065910534455462'),
            balance_raw=Decimal('188153171065910534455462'),
            raw={},
            coin=Coin(
                symbol='SNX',
                name='SNX',
                decimals=18,
                blockchain=Blockchain.OPTIMISM,
                address='0x8700dAec35aF8Ff88c16BdF0418774CB3D7599B4',
                standards=None,
                protocol_id=None,
                info=None,
            ),
            asset_type=AssetType.VESTING,
        ),
    ]



from datetime import datetime
from decimal import Decimal
from functools import lru_cache
from logging import getLogger
from typing import Callable, Dict, Iterable, List, Optional, Tuple

import requests
from eth_typing import ChecksumAddress
from typing_extensions import TypedDict
from web3 import Web3

from blockapi.utils.num import safe_opt_decimal
from blockapi.v2.api.perpetual.perp_abi import rewards_abi
from blockapi.v2.api.web3_utils import easy_call, get_eth_client
from blockapi.v2.base import ApiOptions, BlockchainApi, IBalance
from blockapi.v2.coins import COIN_PERP
from blockapi.v2.models import AssetType, BalanceItem, Blockchain, Coin

logger = getLogger(__name__)


def perp_contract_address(contract_name: str) -> ChecksumAddress:
    """
    Get PERP contract address by name.
    """
    l1 = perp_contracts()['layers']['layer1']
    l1_c = l1['contracts']

    contract_map = {
        'PERP': l1['externalContracts']['perp'],
        'sPERP': l1_c['StakedPerpToken']['address'],
        'staking_rewards': l1_c['PerpStakingRewardNoVesting']['address'],
        'vesting_rewards': l1_c['PerpStakingRewardVesting']['address'],
    }

    if contract_name not in contract_map:
        raise ValueError('Invalid contract name.')

    contract_address = contract_map[contract_name]
    return Web3.toChecksumAddress(contract_address)


@lru_cache()
def perp_contracts() -> dict:
    response = requests.get('https://metadata.perp.exchange/production.json')
    return response.json()


class PerpSnapshot(TypedDict):
    epoch: int
    weekStart: str
    weekEnd: str
    redeemableUntil: int
    totalRewardPerp: str
    hash: str


class PerpReward(TypedDict):
    totalRewardPerp: str
    areaFactor: str


class PerpOffChainStorage:
    """
    Perp Off-chain storage for staking rewards, ... .
    """

    base_url = 'https://s3.amazonaws.com/staking.perp.fi/production'
    snapshot_urls = {
        'immediate': '/snapshot-immediate.json',
        'vesting': '/snapshot-vesting.json',
    }
    epoch_rewards_link = '/{epoch}/{hash}.json'

    @classmethod
    def get_epoch_snapshots(cls, snapshot_type: str) -> List[PerpSnapshot]:
        """
        Load epoch (weekly) immediate or vesting snapshots.
        """
        url = f"{cls.base_url}{cls.snapshot_urls[snapshot_type]}"
        response = requests.get(url)
        snapshots = response.json()

        for epoch, snapshot in snapshots.items():
            snapshot.update({'epoch': int(epoch)})

        return sorted(snapshots.values(), key=lambda i: i['epoch'])

    @classmethod
    @lru_cache(maxsize=128)
    def get_rewards(cls, epoch: int, hash_: str) -> Dict[str, PerpReward]:
        """
        Load perp rewards for whole epoch.
        """
        url = (
            f'{cls.base_url}'
            f'{cls.epoch_rewards_link.format(epoch=epoch, hash=hash_)}'
        )

        result = requests.get(url)

        try:
            return result.json()
        except requests.JSONDecodeError as e:
            logger.error(f"Request parsing to json failed: %s", result.text)
            raise e


class PerpProtocol:
    """
    Documentation how to get rewards and vesting.
    https://docs.google.com/document/d/1E2k8Ad2POodLWWyeklif0tx0-CVy_41_ONAS8n9RN68/edit
    """

    reward_contract = '0xc2a9e84D77f4B534F049b593C282c5c91F24808A'
    vesting_contract = '0x49a4B8431Fc24BE4b22Fb07D1683E2c52bC56088'

    def __init__(self, address, provider="infura"):
        self.address = address
        self.w3 = get_eth_client('mainnet', provider)

    def yield_balances(self) -> Iterable[BalanceItem]:
        """
        Fetch all balances for snx and synth tokens.
        """
        s_claimable = self._fetch_staking_claimable_rewards()
        v_claimable, v_locked = self._fetch_staking_vesting_rewards()
        claimable = s_claimable + v_claimable

        if claimable > Decimal(0):
            yield self._create_balance(AssetType.CLAIMABLE, claimable)

        if v_locked:
            yield self._create_balance(AssetType.VESTING, v_locked)

    def fetch_balances(self) -> List[BalanceItem]:
        return list(self.yield_balances())

    @staticmethod
    def _create_balance(asset_type: AssetType, amount: Decimal) -> BalanceItem:
        return BalanceItem.from_api(
            balance_raw=amount,
            coin=COIN_PERP,
            asset_type=asset_type,
            protocol='perpetual',
            raw={},
        )

    def _fetch_staking_claimable_rewards(self) -> Decimal:
        """
        Fetch claimable staking rewards.
        """
        snapshots = self._fetch_non_claimed_snapshots('immediate', 'staking_rewards')
        return self._get_total_reward(snapshots)

    def _fetch_staking_vesting_rewards(self) -> Tuple[Decimal, Decimal]:
        """
        Fetch staking vesting rewards, both claimable and locked.
        """
        snapshots = self._fetch_non_claimed_snapshots('vesting', 'vesting_rewards')

        current_ts = datetime.utcnow().timestamp()

        claimable = self._get_total_reward(
            snapshots, lambda s: current_ts > s['redeemableUntil']
        )
        locked = self._get_total_reward(
            snapshots, lambda s: current_ts < s['redeemableUntil']
        )

        return claimable, locked

    def _fetch_non_claimed_snapshots(
        self, snapshot_type: str, contract_name: str
    ) -> List[PerpSnapshot]:
        """
        Fetch snapshots and claims for address.
        """
        snapshots = PerpOffChainStorage.get_epoch_snapshots(snapshot_type)
        rewards_contract = self.w3.eth.contract(
            perp_contract_address(contract_name), abi=rewards_abi
        )

        start, end = snapshots[0]['epoch'], snapshots[-1]['epoch']
        claims = easy_call(
            rewards_contract,
            'claimStatus',
            *[self.address, start, end],
        )

        return [snapshot for snapshot, claimed in zip(snapshots, claims) if not claimed]

    def _get_total_reward(
        self,
        snapshots: List[PerpSnapshot],
        snapshot_filter: Optional[Callable] = None,
    ) -> Decimal:
        """
        Get perp total staking reward.
        """
        rewards = (
            self._get_reward(s)
            for s in snapshots
            if snapshot_filter is None or snapshot_filter(s)
        )
        return sum(rewards) or Decimal(0)

    def _get_reward(self, snapshot: PerpSnapshot) -> Decimal:
        """
        Get single epoch's reward.
        """
        rewards = PerpOffChainStorage.get_rewards(snapshot['epoch'], snapshot['hash'])
        return safe_opt_decimal(
            rewards[self.address]['totalRewardPerp']
            if self.address in rewards
            else None
        )


class PerpetualApi(BlockchainApi, IBalance):
    coin = COIN_PERP
    api_options = ApiOptions(
        blockchain=Blockchain.ETHEREUM, base_url=None, rate_limit=0.2
    )

    def __init__(self, api_url: str) -> None:
        super().__init__()
        self.w3 = get_eth_client(api_url)

    def get_balance(self, address: str) -> List[BalanceItem]:
        return PerpProtocol(address).fetch_balances()



import pytest

from blockapi.v2.base import CustomizableBlockchainApi
from blockapi.v2.models import ApiOptions, Blockchain


def test_not_implemented_base_url():
    with pytest.raises(NotImplementedError):
        api = WrongApi()


class WrongApi(CustomizableBlockchainApi):
    api_options = ApiOptions(
        blockchain=Blockchain.ETHEREUM,
        base_url="",
    )



from abc import ABC
from typing import Dict, List, Optional
from urllib.parse import urljoin

from requests import HTTPError, Response, Session

from .models import ApiOptions, BalanceItem, Coin, Pool


class BlockchainApi(ABC):
    """
    General class for handling blockchain API services.
    """

    coin: Coin = NotImplemented
    api_options: ApiOptions = NotImplemented

    # {request_method: request_url}
    supported_requests: Dict[str, str] = {}

    def __init__(self, api_key: Optional[str] = None):
        self.api_key = api_key
        self._session = Session()

    def __del__(self):
        self._session.close()

    def get(
        self,
        request_method: str,
        headers=None,
        **req_args,
    ) -> Dict:
        """
        Call specific request method with params and return raw response.
        """
        url = self._build_request_url(request_method, **req_args)
        response = self._session.get(url, headers=headers)
        return self._check_and_get_from_response(response)

    def _build_request_url(self, request_method: str, **req_args):
        path_url = self.supported_requests.get(request_method)
        if path_url is not None:
            path_url = path_url.format(**req_args)

        return urljoin(self.api_options.base_url, path_url)

    def post(self, request_method=None, body=None, json=None, headers=None):
        """
        Call request using json.
        """
        url = (
            self._build_request_url(request_method)
            if request_method
            else self.api_options.base_url
        )
        response = self._session.post(url, data=body, json=json, headers=headers)
        return self._check_and_get_from_response(response)

    def _check_and_get_from_response(self, response: Response) -> Dict:
        if response.status_code != 200:
            self._raise_from_response(response)
        self._opt_raise_on_other_error(response)

        return response.json()

    @staticmethod
    def _raise_from_response(response: Response) -> None:
        try:
            response.raise_for_status()
        except HTTPError as e:
            raise ApiException(e)

    def _opt_raise_on_other_error(self, response: Response) -> None:
        # implement in child
        return

    def __repr__(self):
        if self.coin is None:
            return self.__class__.__name__

        return f'{self.__class__.__name__}(coin={self.coin.name})'


class CustomizableBlockchainApi(BlockchainApi, ABC):
    """
    Class for handling blockchain API services with customizable base URL,
    e.g. proxy, testnet, RPC services, alternative sources
    """

    API_BASE_URL: str = None

    def __init__(self, base_url: str = None, api_key: Optional[str] = None):
        super().__init__(api_key)
        self.api_options.base_url = base_url or self.API_BASE_URL

        if not self.api_options.base_url:
            raise NotImplementedError(
                'API_BASE_URL is not defined and no base_url was provided'
            )


class IBalance(ABC):
    def get_balance(self, address: str) -> List[BalanceItem]:
        raise NotImplementedError


class IPortfolio(ABC):
    def get_portfolio(self, address: str) -> List[Pool]:
        raise NotImplementedError


class ApiException(Exception):
    pass


class InvalidAddressException(ApiException):
    pass



from blockapi.services import APIKeyMissing, BlockchainAPI


class CryptoIDAPI(BlockchainAPI):
    """
    Multi coins
    API docs: https://chainz.cryptoid.info/api.dws
    Explorer: https://chainz.cryptoid.info
    """

    active = True

    symbol = None
    base_url = 'http://chainz.cryptoid.info'
    rate_limit = 0
    coef = 1
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/{symbol}/api.dws?q=getbalance&a={address}' '&key={api_key}',
    }

    def __init__(self, address, api_key=None):
        if api_key is None:
            raise APIKeyMissing

        super().__init__(address, api_key)

    def get_balance(self):
        response = self.request(
            'get_balance',
            api_key=self.api_key,
            symbol=self.symbol.lower(),
            address=self.address,
        )

        return [{'symbol': self.symbol, 'amount': float(response) * self.coef}]


class CryptoIDLitecoinAPI(CryptoIDAPI):
    symbol = 'LTC'


class CryptoIDDashAPI(CryptoIDAPI):
    symbol = 'DASH'


class CryptoIDGroestlcoinAPI(CryptoIDAPI):
    symbol = 'GRS'



from pytest import mark

from blockapi.api import OntioAPI
from blockapi.test_data import test_addresses


class TestOntioAPI:
    ADDRESS = test_addresses['ONT'][0]

    @mark.vcr()
    def test_get_balance(self):
        api = OntioAPI(self.ADDRESS)
        result = api.get_balance()



from abc import ABC, abstractmethod
from typing import Iterable, List

from blockapi.v2.base import BlockchainApi, IBalance
from blockapi.v2.coins import COIN_BTC, COIN_DOGE, COIN_LTC
from blockapi.v2.models import (
    ApiOptions,
    AssetType,
    BalanceItem,
    Blockchain,
    OperationDirection,
    OperationItem,
    OperationType,
    TransactionItem,
    TransactionStatus,
)


class BlockchairApi(BlockchainApi, IBalance, ABC):
    BASE_URL = 'https://api.blockchair.com'

    @property
    @abstractmethod
    def blockchain(self):
        pass

    @property
    @abstractmethod
    def coin(self):
        pass

    @property
    @abstractmethod
    def name(self):
        pass

    supported_requests = {
        'get_dashboard': '/{name}/dashboards/{address_type}/{address}?limit={limit},0&offset={offset},0',
        'get_txs': '/{name}/dashboards/transactions/{hash_or_hashes}',
    }

    def __init__(self):
        super().__init__()

        self.api_options = ApiOptions(
            blockchain=self.blockchain, base_url=self.BASE_URL, rate_limit=0
        )

    def get_balance(self, address: str) -> List[BalanceItem]:
        dashboard = self._get_dashboard(address)
        return list(self._parse_balances(dashboard))

    def get_transactions(
        self, address: str, *, offset: int = 0, limit: int = 10
    ) -> List[TransactionItem]:
        dashboard = self._get_dashboard(address, offset, limit)
        hashes = dashboard.get('transactions')
        if not hashes:
            return []

        transactions = self.get(
            'get_txs', name=self.name, hash_or_hashes=','.join(hashes)
        )

        data = transactions.get('data')
        if not data:
            return []

        return list(self._parse_transactions(address, data))

    def _get_dashboard(self, address, offset: int = 0, limit: int = 10):
        address_type = self._get_address_type(address)
        raw_dashboard = self.get(
            'get_dashboard',
            address=address,
            symbol=self.coin.symbol,
            address_type=address_type,
            name=self.name,
            offset=offset,
            limit=limit,
        )

        return self._parse_dashboard(raw_dashboard, address_type)

    @staticmethod
    def _get_address_type(address):
        is_xpub = any(address.startswith(p) for p in ['xpub', 'ypub', 'zpub'])

        return 'xpub' if is_xpub else 'address'

    @staticmethod
    def _parse_dashboard(dashboard: dict, address_type: str):
        data = dashboard.get('data')
        if not data:
            return

        item = list(data.values())[0]
        if address_type == 'address':
            adr = item.get('address')
            if not adr.get('type'):
                return

        return item

    def _parse_balances(self, dashboard: dict) -> Iterable[BalanceItem]:
        addr = dashboard.get('address')
        balance = addr.get('balance')

        yield BalanceItem.from_api(
            balance_raw=balance,
            coin=self.coin,
            asset_type=AssetType.AVAILABLE,
            raw=dashboard,
        )

    def _parse_transactions(
        self, address: str, data: dict
    ) -> Iterable[TransactionItem]:
        for tx in data.values():
            parsed = self._parse_transaction(address, tx)
            if parsed:
                yield parsed

    def _parse_transaction(self, address: str, tx: dict) -> TransactionItem:
        incoming = next(
            (i for i in tx.get('inputs') if address == i['recipient']), None
        )
        outgoing = next(
            (o for o in tx.get('outputs') if address == o['recipient']), None
        )

        tx_data = tx.get('transaction')

        if incoming:
            return TransactionItem.from_api(
                fee_raw=tx_data.get('fee'),
                coin=self.coin,
                date=tx.get('time'),
                hash=incoming.get('hash'),
                status=TransactionStatus.CONFIRMED,
                operations=[
                    OperationItem.from_api(
                        amount_raw=incoming.get('value'),
                        coin=self.coin,
                        to_address=address,
                        from_address=(
                            tx['inputs'][0]['recipient']
                            if tx_data['input_count'] == 1
                            else 'multiple'
                        ),
                        hash=tx_data.get('hash'),
                        type=OperationType.TRANSACTION,
                        direction=OperationDirection.INCOMING,
                        raw={},
                        confirmed=None,
                    )
                ],
                raw=tx,
            )

        if outgoing:
            return TransactionItem.from_api(
                fee_raw=tx.get('fee'),
                coin=self.coin,
                date=tx.get('created'),
                hash=outgoing.get('hash'),
                status=TransactionStatus.CONFIRMED,
                operations=[
                    OperationItem.from_api(
                        amount_raw=outgoing.get('value'),
                        coin=self.coin,
                        from_address=address,
                        to_address=(
                            tx['outputs'][0]['recipient']
                            if tx_data['output_count'] == 1
                            else 'multiple'
                        ),
                        hash=tx_data.get('hash'),
                        type=OperationType.TRANSACTION,
                        direction=OperationDirection.OUTGOING,
                        raw={},
                        confirmed=None,
                    )
                ],
                raw=tx,
            )


class BlockchairBitcoinApi(BlockchairApi):
    coin = COIN_BTC
    blockchain = Blockchain.BITCOIN
    name = 'bitcoin'


class BlockchairDogecoinApi(BlockchairApi):
    coin = COIN_DOGE
    blockchain = Blockchain.DOGECHAIN
    name = 'dogecoin'


class BlockchairLitecoinApi(BlockchairApi):
    coin = COIN_LTC
    blockchain = Blockchain.LITECOIN
    name = 'litecoin'



from decimal import Decimal, InvalidOperation
from numbers import Number
from typing import Optional, Union

SupportsNumber = Union[str, Number]


def to_int(number: Union[int, str]):
    return int(number)


def to_decimal(number: Union[int, float, str, Decimal]) -> Decimal:
    """
    Try to convert any number to Decimal.
    Suited mostly for floats:
    >>> Decimal(1.01)
    Decimal('1.0100000000000000088817841970012523233890533447265625')
    >>> Decimal('1.01')
    Decimal('1.01')
    """
    if isinstance(number, Decimal):
        return number
    elif isinstance(number, float):
        return Decimal(str(number))
    elif isinstance(number, (int, str)):
        return Decimal(number)
    else:
        raise TypeError(f'Type {type(number)} is not supported.')


def raw_to_decimals(raw: Union[int, str], decimals: Union[int, str]) -> Decimal:
    """
    Conversion of raw (gwei/satoshi/...) format to decimals.
    """
    raw_ = to_decimal(raw)
    decimals_ = to_decimal(decimals)
    res = raw_ * pow(10, -decimals_)
    return remove_exponent(res)


def decimals_to_raw(amount: Union[int, str], decimals: Union[int, str]) -> Decimal:
    """
    Convert decimal value to raw format
    """
    amount_ = to_decimal(amount)
    decimals_ = to_decimal(decimals)
    res = amount_ * pow(10, decimals_)
    return remove_exponent(res)


def remove_exponent(d: Decimal) -> Decimal:
    """
    https://docs.python.org/3/library/decimal.html#decimal-faq
    Remove the exponent and trailing zeroes, losing significance,
    but keeping the value unchanged.
    """
    try:
        return d.quantize(Decimal(1)) if d == d.to_integral() else d.normalize()
    except InvalidOperation:
        return d


def safe_opt_decimal(obj: Optional[SupportsNumber]) -> Decimal:
    """
    This function is primarily for parsing 3rd party APIs that sometimes,
    despite stating in their docs that *number* is returned, may return null.

    >>> safe_opt_decimal(None)
    Decimal('0')
    """
    if obj is None:
        return Decimal('0')
    else:
        return to_decimal(obj)


def safe_decimal(number: SupportsNumber) -> Decimal:
    """
    Try to convert any number to Decimal.
    Suited mostly for floats:
    >>> Decimal(1.01)
    Decimal('1.0100000000000000088817841970012523233890533447265625')
    >>> Decimal('1.01')
    Decimal('1.01')
    """
    if isinstance(number, Decimal):
        return number
    if isinstance(number, float):
        return Decimal(str(number))
    elif isinstance(number, (int, str)):
        return Decimal(number)
    else:
        raise TypeError(f'Type {type(number)} is not supported.')



import json
import os
from typing import List


def read_json_file(file_name: str) -> List:
    json_path = os.path.abspath(os.path.join(os.path.dirname(__file__), file_name))
    with open(json_path, encoding="utf-8") as f:
        return json.load(f)


def read_file(file_name: str) -> str:
    json_path = os.path.abspath(os.path.join(os.path.dirname(__file__), file_name))
    with open(json_path, encoding="utf-8") as f:
        return f.read()



from decimal import Decimal

import pytest

from blockapi.test.v2.api.conftest import read_file
from blockapi.v2.api.blockchair import BlockchairLitecoinApi

# noinspection SpellCheckingInspection
ltc_test_address = 'M8T1B2Z97gVdvmfkQcAtYbEepune1tzGua'


def test_fetch_balances(requests_mock, ltc_balance_response):
    requests_mock.get(
        f'https://api.blockchair.com/litecoin/dashboards/address/{ltc_test_address}?limit=10,0&offset=0,0',
        text=ltc_balance_response,
    )

    api = BlockchairLitecoinApi()
    balances = api.get_balance(ltc_test_address)
    assert len(balances) == 1
    assert balances[0].balance == Decimal('2524667.31490873')


def test_fetch_transactions(
    requests_mock, ltc_balance_response, ltc_transactions_response
):
    requests_mock.get(
        f'https://api.blockchair.com/litecoin/dashboards/address/{ltc_test_address}?limit=3,0&offset=0,0',
        text=ltc_balance_response,
    )

    # noinspection SpellCheckingInspection
    addresses = [
        "6b6065edce69d513d81bd2960562062577e1678ee07a24f54d2c5bb660b8ab22",
    ]

    requests_mock.get(
        'https://api.blockchair.com/litecoin/dashboards/transactions/'
        + ','.join(addresses),
        text=ltc_transactions_response,
    )

    api = BlockchairLitecoinApi()

    txs = api.get_transactions(ltc_test_address, limit=3)
    assert len(txs) == 1
    assert txs[0].fee == Decimal('0')
    assert len(txs[0].operations) == 1
    assert txs[0].operations[0].amount == Decimal('0.0002')


@pytest.fixture()
def ltc_transactions_response():
    return read_file('data/blockchair_ltc_transaction_response.json')


@pytest.fixture
def ltc_balance_response():
    return read_file('data/blockchair_ltc_balance_response.json')



import logging
from abc import ABCMeta, abstractmethod
from typing import Dict

from eth_utils import to_checksum_address

from blockapi.v2.base import BlockchainApi, IBalance
from blockapi.v2.models import BalanceItem, Coin, CoinInfo

logger = logging.getLogger(__name__)


class CovalentApiBase(BlockchainApi, IBalance, metaclass=ABCMeta):
    """
    API docs: https://www.covalenthq.com/docs/api/
    """

    API_BASE_URL = 'https://api.covalenthq.com/v1'
    API_BASE_RATE_LIMIT = 0.2

    @property
    @abstractmethod
    def api_options(self):
        pass

    @property
    @abstractmethod
    def CHAIN_ID(self):
        pass

    @property
    @abstractmethod
    def coin(self):
        pass

    supported_requests = {
        'get_balance': '/v1/{chain_id}/address/{address}/balances_v2/'
    }

    def __init__(self, api_key: str):
        super().__init__(api_key)

        # Set http basic auth for requests.
        self._session.auth = (api_key, "")

    def get_balance(self, address: str) -> [BalanceItem]:
        response = self.get('get_balance', chain_id=self.CHAIN_ID, address=address)

        return self._parse_items(response)

    @staticmethod
    def to_checksum_address(address: str):
        try:
            return to_checksum_address(address)
        except ValueError as e:
            logger.exception(e)
            return address

    def _parse_items(self, response: Dict) -> [BalanceItem]:
        try:
            raw_balances = response['data']['items']
        except KeyError as e:
            logger.exception(e)
            return []

        balances = []
        for raw_balance in raw_balances:
            if raw_balance.get('balance') is None or raw_balance['balance'] == 0:
                logger.debug(
                    "Skipping coin: '%s' - balance is zero.",
                    raw_balance.get("contract_name"),
                )
                continue

            coin_symbol = raw_balance.get('contract_ticker_symbol')
            if coin_symbol == self.coin.symbol:
                # Native coin for given blockchain.
                coin = self.coin
            else:
                coin = Coin.from_api(
                    symbol=raw_balance.get('contract_ticker_symbol'),
                    name=raw_balance.get('contract_name'),
                    decimals=raw_balance.get('contract_decimals', 0),
                    blockchain=self.api_options.blockchain,
                    address=self.to_checksum_address(
                        raw_balance.get('contract_address')
                    ),
                    standards=raw_balance.get("supports_erc", []),
                    info=CoinInfo(logo_url=raw_balance.get("logo_url")),
                )

            balances.append(
                BalanceItem.from_api(
                    balance_raw=raw_balance.get('balance'),
                    coin=coin,
                    last_updated=raw_balance.get('last_transferred_at'),
                    raw=raw_balance,
                )
            )

        return balances



from decimal import Decimal

from blockapi.v2.coins import COIN_ETH
from blockapi.v2.models import BalanceItem, Blockchain, Protocol


def test_create_protocol_from_api():
    protocol = Protocol.from_api(
        protocol_id="0xmons",
        chain=Blockchain.ETHEREUM,
        name="0xmons_name",
        user_deposit=1234.5,
        site_url="https://0xmons.xyz",
        logo_url="https://static.debank.com/image/project/logo_url/0xmons/125fd50693dcfb1a1c4a720fb27f017d.png",
        has_supported_portfolio=True,
    )

    assert protocol.protocol_id == '0xmons'
    assert protocol.chain == Blockchain.ETHEREUM
    assert protocol.name == '0xmons_name'
    assert protocol.site_url == 'https://0xmons.xyz'
    assert (
        protocol.logo_url == 'https://static.debank.com/image/project/logo_url/0xmons/'
        '125fd50693dcfb1a1c4a720fb27f017d.png'
    )
    assert protocol.has_supported_portfolio is True
    assert protocol.user_deposit == Decimal(1234.5)


def test_balance_has_protocol():
    balance = BalanceItem.from_api(balance_raw="3", coin=COIN_ETH, raw={})
    assert balance.protocol is None


def test_balance_protocol_can_be_set():
    protocol = Protocol.from_api(
        protocol_id="x", chain=Blockchain.ETHEREUM, name="X", user_deposit=4
    )
    balance = BalanceItem.from_api(
        balance_raw="3", coin=COIN_ETH, raw={}, protocol=protocol
    )
    assert balance.protocol == protocol



import dateutil.parser

from blockapi.services import APIError, BlockchainAPI, set_default_args_values


class TzscanAPI(BlockchainAPI):
    """
    Tezos
    API docs: https://tzscan.io/api
        Explorer: https://tzscan.io
    """

    # inaccurate results for balances, use on your own risk
    active = False

    symbol = 'XTZ'
    base_url = None  # endpoint is created in runtime
    rate_limit = 0
    coef = 1e-6
    max_items_per_page = 50
    page_offset_step = 1

    supported_requests = {
        'get_balance': '/v3/balance_from_balance_updates/{address}',
        'get_operations': '/v3/operations/{address}'
        '?type={type}&p={page_offset}&number={number}',
        'get_rewards': '/v3/rewards_split_cycles/{address}'
        '?p={page_offset}&number={number}',
        'get_rewards_split': '/v2/rewards_split/{address}'
        '?cycle={cycle}&number=50&p={page}',
        'get_bakings': '/v3/cycle_bakings/{address}' '?p={page_offset}&number={number}',
        'get_endorsements': '/v3/cycle_endorsements/{address}'
        '?p={page_offset}&number={number}',
    }

    def __init__(self, address, *args, **kwargs):
        super().__init__(address)
        # this API is very unstable, add some functionality to prevent fails
        self._base_url_temp = 'https://api{num}.dunscan.io'  # num is 1-6
        self._api_nums = [1, 2, 3, 4, 5, 6]

    def get_balance(self):
        balance = self._safe_request('get_balance', address=self.address)
        return [
            {'symbol': self.symbol, 'amount': float(balance['spendable']) * self.coef}
        ]

    def get_rewards(self, offset=0, limit=50):
        rewards = self._safe_request(
            'get_rewards', address=self.address, page_offset=offset, number=limit
        )

        return [
            self.get_reward_details(reward, int(reward['cycle'])) for reward in rewards
        ]

    def get_reward_details(self, reward_v3, cycle=None):
        reward = self._safe_request(
            'get_rewards_split', address=self.address, cycle=cycle, page=0
        )

        total_staking_balance = float(reward['delegate_staking_balance'])
        total_rewards = (
            float(reward['blocks_rewards'])
            + float(reward['endorsements_rewards'])
            + float(reward['fees'])
            + float(reward['future_blocks_rewards'])
            + float(reward['future_endorsements_rewards'])
            + float(reward['gain_from_denounciation_baking'])
            + float(reward['revelation_rewards'])
            + float(reward['gain_from_denounciation_endorsement'])
            - float(reward['lost_deposit_from_denounciation_baking'])
            - float(reward['lost_fees_denounciation_baking'])
            - float(reward['lost_rewards_denounciation_baking'])
            - float(reward['lost_deposit_from_denounciation_endorsement'])
            - float(reward['lost_fees_denounciation_endorsement'])
            - float(reward['lost_rewards_denounciation_endorsement'])
            - float(reward['lost_revelation_rewards'])
            - float(reward['lost_revelation_fees'])
        )

        total_staking_balance = round(float(total_staking_balance) / 1000000, 6)
        total_delegators = int(reward['delegators_nb'])
        pages = total_delegators / 50

        delegations = []
        delegators = 0
        page = 0

        while True:
            for del_balance in reward['delegators_balance']:
                delegator_address = del_balance[0]['tz']
                delegator_balance = int(del_balance[1]) * self.coef
                share = float(delegator_balance) / total_staking_balance
                delegations.append(
                    {
                        'address': delegator_address,
                        'balance': delegator_balance,
                        'share': share * 100,
                        'rewards_share': share * (total_rewards * self.coef),
                    }
                )
                delegators += 1

            page += 1
            if page < pages:
                reward = self._safe_request(
                    'get_rewards_split', address=self.address, cycle=cycle, page=page
                )
            else:
                break

        return {
            'staking_balance': total_staking_balance,
            'cycle': cycle,
            'status': reward_v3['status']['status'],
            'delegators': delegators,
            'delegations': delegations,
            'end_rewards': float(reward['endorsements_rewards']) * self.coef,
            'blocks_rewards': float(reward['blocks_rewards']) * self.coef,
            'extra_rewards': float(reward['revelation_rewards']) * self.coef,
            'losses': float(reward['lost_revelation_rewards']) * self.coef,
            'fees': float(reward['fees']) * self.coef,
        }

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        return self._get_operations('Transaction', self.parse_tx, offset, limit)

    def get_activations(self, offset=None, limit=None):
        return self._get_operations('Activation', self.parse_activation, offset, limit)

    def get_originations(self, offset=None, limit=None):
        return self._get_operations(
            'Origination', self.parse_origination, offset, limit
        )

    def get_delegations(self, offset=None, limit=None):
        return self._get_operations('Delegation', self.parse_delegation, offset, limit)

    @set_default_args_values
    def _get_operations(self, op_type, parse, offset=None, limit=None):
        """Get all operations by type
        @op_type in [Transaction, Origination, Delegation, Activation, ...]"""

        operations = self._safe_request(
            'get_operations',
            address=self.address,
            type=op_type,
            page_offset=offset,
            number=limit,
        )

        parsed_txs = []
        for tx in operations:
            parsed_txs += parse(tx)
        return parsed_txs

    def parse_tx(self, tx):
        # tezos can have multiple txs in single tx (with one hash)
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append(
                {
                    'date': dateutil.parser.parse(op['timestamp']),
                    'from_address': op['src']['tz'],
                    'to_address': op['destination']['tz'],
                    'amount': float(op['amount']) * self.coef,
                    'fee': None
                    if float(op['fee']) == -1
                    else float(op['fee']) * self.coef,
                    'gas_limit': None
                    if int(op['gas_limit']) == -1
                    else int(op['gas_limit']),
                    'hash': tx['hash'],
                    'confirmed': None,
                    'is_error': op['failed'],
                    'type': 'internal' if op['internal'] else 'normal',
                    'kind': op['kind'].lower(),
                    'direction': 'outgoing'
                    if self.address == op['src']['tz']
                    else 'incoming',
                    'raw': tx,
                }
            )
            return parsed

    def parse_delegation(self, tx):
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append(
                {
                    'date': dateutil.parser.parse(op['timestamp']),
                    'source_address': op['src']['tz'],
                    'delegate': op['delegate']['tz'],
                    'fee': None if op['fee'] == -1 else op['fee'] * self.coef,
                    'gas_limit': None
                    if int(op['gas_limit']) == -1
                    else int(op['gas_limit']),
                    'hash': tx['hash'],
                    'is_error': op['failed'],
                    'type': 'internal' if op['internal'] else 'normal',
                    'kind': op['kind'].lower(),
                    'raw': tx,
                }
            )
            return parsed

    @staticmethod
    def parse_activation(tx):
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append(
                {
                    'date': dateutil.parser.parse(op['timestamp']),
                    'secret': op['secret'],
                    'balance': op['balance'],
                    'hash': tx['hash'],
                    'kind': op['kind'].lower(),
                    'raw': tx,
                }
            )
            return parsed

    def parse_origination(self, tx):
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append(
                {
                    'date': dateutil.parser.parse(op['timestamp']),
                    'originator': op['src']['tz'],
                    'manager': op['managerPubkey']['tz'],
                    'balance': op['balance'],
                    'spendable': op['spendable'],
                    'delegatable': op['delegatable'],
                    'delegate': op['delegate']['tz'],
                    'delegate_alias': op['delegate']['alias'],
                    'burnt': op['burn_tez'] * self.coef,
                    'fee': None if op['fee'] == -1 else op['fee'] * self.coef,
                    'gas_limit': None
                    if int(op['gas_limit']) == -1
                    else int(op['gas_limit']),
                    'hash': tx['hash'],
                    'is_error': op['failed'],
                    'type': 'internal' if op['internal'] else 'normal',
                    'kind': op['kind'].lower(),
                    'raw': tx,
                }
            )
            return parsed

    @set_default_args_values
    def get_endorsements(self, offset=None, limit=None):
        ends = self._safe_request(
            'get_endorsements', address=self.address, page_offset=offset, number=limit
        )
        return [self.parse_endorsement(e) for e in ends]

    def parse_endorsement(self, e):
        return {
            'cycle': int(e['cycle']),
            'depth': int(e['depth']),
            'num_slots': {
                'all': int(e['slots']['count_all']),
                'miss': int(e['slots']['count_miss']),
                'steal': int(e['slots']['count_steal']),
            },
            'fee': int(e['tez']['fee']) * self.coef,
            'reward': int(e['tez']['reward']) * self.coef,
            'deposit': int(e['tez']['deposit']) * self.coef,
            'priority': float(e['priority']),
        }

    @set_default_args_values
    def get_bakings(self, offset=None, limit=None):
        baks = self._safe_request(
            'get_bakings', address=self.address, page_offset=offset, number=limit
        )
        return [self.parse_baking(b) for b in baks]

    def parse_baking(self, b):
        return {
            'cycle': int(b['cycle']),
            'depth': int(b['depth']),
            'num_blocks': {
                'all': int(b['count']['count_all']),
                'miss': int(b['count']['count_miss']),
                'steal': int(b['count']['count_steal']),
            },
            'fee': int(b['tez']['fee']) * self.coef,
            'reward': int(b['tez']['reward']) * self.coef,
            'deposit': int(b['tez']['deposit']) * self.coef,
            'priority': float(b['priority']),
            'bake_time': int(b['bake_time']),
        }

    def _safe_request(self, method, **params):
        api_nums = self._api_nums.copy()
        ex = Exception('Runtime error')

        while api_nums:
            try:
                api_num = api_nums.pop(0)
                self.base_url = self._base_url_temp.format(num=api_num)
                response = self.request(method, **params)
            except APIError as e:
                ex = e
                continue
            else:
                return response

        # when all endpoints are checked and none of them is valid, raise exc
        raise ex



import logging
from typing import Optional, Union

from blockapi.v2.models import Blockchain

logger = logging.getLogger(__name__)

DEBANK_BLOCKCHAINS_MAP = {
    'arb': Blockchain.ARBITRUM,
    'avax': Blockchain.AVALANCHE,
    'bsc': Blockchain.BINANCE_SMART_CHAIN,
    'btt': Blockchain.BIT_TORRENT,
    'cro': Blockchain.CRONOS,
    'dfk': Blockchain.DEFI_KINGDOMS,
    'eth': Blockchain.ETHEREUM,
    'evmos': Blockchain.COSMOS,
    'ftm': Blockchain.FANTOM,
    'hmy': Blockchain.HARMONY,
    'iotx': Blockchain.IOTEX,
    'kcc': Blockchain.KUCOIN,
    'klay': Blockchain.KLAY_TOKEN,
    'matic': Blockchain.POLYGON,
    'metis': Blockchain.METIS_ANDROMEDA,
    'movr': Blockchain.MOONRIVER,
    'mobm': Blockchain.MOONBEAM,
    'op': Blockchain.OPTIMISM,
    'sbch': Blockchain.SMART_BITCOIN_CASH,
    'sdn': Blockchain.SHIDEN_NETWORK,
    'sgb': Blockchain.SONGBIRD,
    'tlos': Blockchain.TELOS,
    'wan': Blockchain.WANCHAIN,
}

COINGECKO_BLOCKCHAINS_MAP = {
    'defi-kingdom-blockchain': Blockchain.DEFI_KINGDOMS,
    'defi-kingdoms-blockchain': Blockchain.DEFI_KINGDOMS,
    'harmony-shard-0': Blockchain.HARMONY,
    'kucoin-community-chain': Blockchain.KUCOIN,
    'polygon-pos': Blockchain.POLYGON,
    'oec': Blockchain.OKT,
    'shibchain': Blockchain.SHIBA_CHAIN,
    'shiden network': Blockchain.SHIDEN_NETWORK,
    'smartbch': Blockchain.SMART_BITCOIN_CASH,
    'thorchain': Blockchain.THOR,
}

CHAIN_ID_BLOCKCHAINS_MAP = {
    '1': Blockchain.ETHEREUM,
    '2': Blockchain.EXPANSE_NETWORK,
    '3': None,  # testnet - Ropsten
    '4': None,  # testnet,
    '5': None,  # testnet,
    '6': None,  # testnet - Ethereum Classic Testnet Kotti
    '7': Blockchain.THAI_CHAIN,
    '8': Blockchain.UBIQ,
    '10': Blockchain.OPTIMISM,
    '12': None,  # testnet - Metadium Testnet
    '14': Blockchain.FLARE,
    '16': None,  # testnet - Flare Testnet Coston
    '17': Blockchain.THAI_CHAIN_2_THAI_FI,
    '19': Blockchain.SONGBIRD,
    '20': Blockchain.ELASTOS,
    '21': None,  # testnet - ELA-ETH-Sidechain Testnet
    '22': Blockchain.ELA_DID_SIDECHAIN,
    '23': None,  # testnet - ELA-DID-Sidechain Testnet
    '24': Blockchain.DITHEREUM,
    '25': Blockchain.CRONOS,
    '26': None,  # testnet - Genesis L1 testnet
    '27': Blockchain.SHIBA_CHAIN,
    '28': None,  # testnet
    '29': Blockchain.GENESIS_L1,
    '30': Blockchain.ROOTSTOCK,
    '31': None,  # testnet - rsk
    '32': None,  # testnet - good data
    '35': Blockchain.TBWG_CHAIN,
    '36': Blockchain.DXCHAIN,
    '37': Blockchain.SEED_COIN_NETWORK,
    '38': Blockchain.VALORBIT,
    '40': Blockchain.TELOS,
    '44': Blockchain.DARWINIA_CRAB_NETWORK,
    '56': Blockchain.BINANCE_SMART_CHAIN,
    '57': Blockchain.SYSCOIN,
    '58': Blockchain.ONTOLOGY,
    '61': Blockchain.ETHEREUM_CLASSIC,
    '66': Blockchain.OKT,
    '69': None,  # testnet
    '97': None,  # testnet
    '100': Blockchain.GNOSIS,
    '106': Blockchain.VELAS,
    '122': Blockchain.FUSE,
    '128': Blockchain.HUOBI_TOKEN,
    '137': Blockchain.POLYGON,
    '250': Blockchain.FANTOM,
    '256': None,  # testnet
    '288': Blockchain.BOBA,
    '321': Blockchain.KCC,
    '336': Blockchain.SHIDEN_NETWORK,
    '416': Blockchain.SX_NETWORK,
    '592': Blockchain.ASTAR,
    '1024': Blockchain.CLV_PARACHAIN,
    '1030': Blockchain.CONFLUX,
    '1088': Blockchain.METIS_ANDROMEDA,
    '1284': Blockchain.MOONBEAM,
    '1285': Blockchain.MOONRIVER,
    '2000': Blockchain.DOGECHAIN,
    '2001': Blockchain.MILKOMEDA_C1,
    '2222': Blockchain.KAVA,
    '4689': Blockchain.IOTEX,
    '8217': Blockchain.KLAYTN_CYPRESS,
    '9001': Blockchain.EVMOS,
    # "16350": None, #TODO: Not found
    '32659': Blockchain.FUSION_NETWORK,
    '42161': Blockchain.ARBITRUM,
    '42170': Blockchain.ARBITRUM_NOVA,
    '42220': Blockchain.CELO,
    '42262': Blockchain.EMERALD_PARATIME,
    '43114': Blockchain.AVALANCHE,
    '47805': Blockchain.REI_NETWORK,
    '53935': Blockchain.DFK_CHAIN,
    '71402': Blockchain.GODWOKEN,
    # "73772":None, #TODO: not found
    '210425': Blockchain.PLAT_ON,
    '256256': Blockchain.CMP,
    # "12340001": None, #TODO: not found
    '1313161554': Blockchain.AURORA,
    '1666600000': Blockchain.HARMONY,
}

RANGO_BLOCKCHAINS_MAP = {
    'avax_cchain': Blockchain.AVALANCHE,
    'bnb': Blockchain.BINANCECOIN,
    'bsc': Blockchain.BINANCE_SMART_CHAIN,
    'btc': Blockchain.BITCOIN,
    'crypto_org': Blockchain.CRYPTO_ORG,
    'doge': Blockchain.DOGECHAIN,
    'eth': Blockchain.ETHEREUM,
    'ltc': Blockchain.LITECOIN,
}

WORMHOLE_BLOCKCHAINS_MAP = {
    'eth': Blockchain.ETHEREUM,
    'sol': Blockchain.SOLANA,
    'bsc': Blockchain.BINANCE_SMART_CHAIN,
    'matic': Blockchain.POLYGON,
    'avax': Blockchain.AVALANCHE,
    'ftm': Blockchain.FANTOM,
}


def _get_chain_mapping(
    chain: Optional[Union[str, int]], source: str, mapping: dict[str, Blockchain]
) -> Optional[Blockchain]:
    if not chain:
        return None

    chain_lower = chain.lower() if hasattr(chain, 'lower') else str(chain)

    blockchain = mapping.get(chain_lower)
    if blockchain:
        return blockchain

    if chain_lower in Blockchain.__members__.values():
        return Blockchain(chain_lower)

    logger.warning(f'Cannot convert chain id "{chain}" to Blockchain for {source}')
    return None


def get_blockchain_from_debank_chain(chain: Optional[str]) -> Optional[Blockchain]:
    return _get_chain_mapping(chain, 'DeBank', DEBANK_BLOCKCHAINS_MAP)


def get_blockchain_from_coingecko_chain(chain: Optional[str]) -> Optional[Blockchain]:
    return _get_chain_mapping(chain, 'CoinGecko', COINGECKO_BLOCKCHAINS_MAP)


def get_blockchain_from_chain_id(
    chain: Optional[Union[str, int]]
) -> Optional[Blockchain]:
    return _get_chain_mapping(chain, 'Chain ID', CHAIN_ID_BLOCKCHAINS_MAP)


def get_blockchain_from_rango_chain(chain: Optional[str]) -> Optional[Blockchain]:
    return _get_chain_mapping(chain, 'Rango', RANGO_BLOCKCHAINS_MAP)


def get_blockchain_from_wormhole_chain(chain: Optional[str]) -> Optional[Blockchain]:
    return _get_chain_mapping(chain, 'Wormhole', WORMHOLE_BLOCKCHAINS_MAP)



from decimal import Decimal

from blockapi import (
    check_address_valid,
    get_all_supported_coins,
    get_api_classes_for_coin,
)
from blockapi.test_data import get_test_api_key, test_addresses


class TestBlockApiProviders:
    def __init__(self, data=None, methods=None):
        """State what currencies should be tested on which addresses
        with which expected values

        Args:
            data (list): list of dicts with structure
            {
                "currency": (str),  # Should be valid symbol
                "address": (str),
                "value": (Decimal)
            }
            methods (list): what methods should be tested, if None
                get_balance is tested
        """

        self.currencies = get_all_supported_coins()
        self.data = data if data is not None else []
        self.methods = methods if methods else ['get_balance']
        self.results = {m: [] for m in self.methods}
        self.diagnostic = None
        self.error_api = []

    def run_all(self):
        """Auto run of all APIs using test data"""
        data = []

        # Create test sample
        for currency in self.currencies:
            if not test_addresses.get(currency):
                continue

            data.append(
                {
                    'currency': currency,
                    'address': test_addresses[currency][0],
                    'value': '0',
                }
            )

        self.data = data
        self.run()

    def run(self):
        """Runs tests for given data"""

        for currency in self.data:
            classes = get_api_classes_for_coin(symbol=currency['currency'])

            for class_ in classes:
                # try to load api_key
                api_key = get_test_api_key(class_.__name__)
                try:
                    class_instance = class_(
                        address=currency['address'], api_key=api_key
                    )
                except Exception as e:
                    self.error_api.append(
                        {'api': class_.__name__, 'error': 'Init error', 'message': e}
                    )
                    continue

                for method in self.methods:
                    try:
                        self.results[method].append(
                            {
                                'result': getattr(class_instance, method)(),
                                'api': class_.__name__,
                                'currency': currency['currency'],
                                'success': True,
                            }
                        )
                    except Exception as e:
                        self.error_api.append(
                            {
                                'api': class_.__name__,
                                'error': f'{method} error',
                                'message': e,
                            }
                        )

    def get_diagnostic(self, method=None):
        """Returns info about results"""

        if method is None:
            method = 'get_balance'

        if not self.results:
            return 'Run "run" or "run_all" first.'

        if method not in self.methods:
            return 'Choose valid method.'

        return {
            'api_with_error': [a['api'] for a in self.error_api],
            'api_with_result_none': [
                r['api'] for r in self.results[method] if r['result'] is None
            ],
            'api_with_structure_error': [
                r['api']
                for r in self.results[method]
                if not isinstance(r['result'], list)
            ],
            'api_with_results': {
                c: {
                    'results': [
                        r
                        for r in self.results[method]
                        if r['result'] is not None
                        and r['currency'] == c
                        and isinstance(r['result'], list)
                    ],
                    'mean': self._mean(method=method, currency=c),
                }
                for c in self.currencies
            },
        }

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, _data):
        valid_data = []
        not_valid_data = []
        for d in _data:
            if d['currency'] not in self.currencies:
                not_valid_data.append({'message': 'Symbol not valid', 'data': d})
                continue

            if not check_address_valid(d['currency'], d['address']):
                not_valid_data.append(
                    {'message': 'Check address not successful', 'data': d}
                )
                continue

            try:
                d['value'] = Decimal(d['value'])
            except ValueError:
                not_valid_data.append({'message': 'Value is not a number', 'data': d})
                continue

            valid_data.append(d)

        self._data = valid_data

    def _mean(self, method, currency):
        records = [
            r['result']
            for r in self.results[method]
            if r['result'] is not None
            and r['currency'] == currency
            and isinstance(r['result'], list)
        ]

        sum_ = sum(
            [
                next((i['amount'] for i in r if i['symbol'] == currency), 0)
                for r in records
            ]
        )
        return sum_ / len(records) if records else 0



from datetime import datetime
from typing import Union

from dateutil.parser import parse as parse_date


def parse_dt(dt: Union[str, int, float]) -> datetime:
    """
    Convert datetime from string or timestamp into `datetime.datetime`
    """
    if isinstance(dt, str):
        return parse_date(dt)
    elif isinstance(dt, int):
        return datetime.fromtimestamp(dt)
    elif isinstance(dt, float):
        return datetime.fromtimestamp(dt)
    else:
        raise TypeError(f'Type {type(dt)} is not supported.')



from blockapi.v2.api.covalenth.base import CovalentApiBase
from blockapi.v2.base import ApiOptions
from blockapi.v2.coins import COIN_RSK
from blockapi.v2.models import Blockchain


class RskCovalentApi(CovalentApiBase):
    CHAIN_ID = 30

    api_options = ApiOptions(
        blockchain=Blockchain.ROOTSTOCK,
        base_url=CovalentApiBase.API_BASE_URL,
        rate_limit=CovalentApiBase.API_BASE_RATE_LIMIT,
    )

    coin = COIN_RSK



import json
from decimal import Decimal

from blockapi.services import APIError, BlockchainAPI
from blockapi.utils.num import to_decimal


class SolanaApi(BlockchainAPI):
    """
    Solana RPC
    API docs: https://docs.solana.com/apps/jsonrpc-api
    """

    symbol = 'SOL'
    base_url = 'https://api.mainnet-beta.solana.com/'
    rate_limit = None
    coef = Decimal('1e-9')
    start_offset = 0
    max_items_per_page = 1000
    page_offset_step = 1

    # follow used pattern even though this API uses POST requests
    supported_requests = {'get_balance': '', 'get_txs_signatures': '', 'get_tx': ''}

    token_program_id = 'TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA'

    # mainnet tokens
    # https://github.com/solana-labs/token-list/blob/main/src/tokens/solana.tokenlist.json
    tokens = {
        'SRMuApVNdxXokk5GT7XD5cUUgXMBCoAz2LHeuAoKWRt': {
            'name': 'Serum',
            'symbol': 'SRM',
        },
        'MSRMcoVyrFxnSgo5uXwone5SKcGhT1KEJMFEkMEWf9L': {
            'name': 'MegaSerum',
            'symbol': 'MSRM',
        },
        '9n4nbM75f5Ui33ZbPYXn59EwSgE8CGsHtAeTH5YFeJ9E': {
            'symbol': 'BTC',
            'name': 'Wrapped Bitcoin',
        },
        '2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk': {
            'symbol': 'ETH',
            'name': 'Wrapped Ethereum',
        },
        'BQcdHdAQW1hczDbBi9hiegXAR7A98Q9jx3X3iBBBDiq4': {
            'symbol': 'USDT',
            'name': 'Wrapped USDT',
        },
        'BXXkv6z8ykpG1yuvUDPgh732wzVHB69RnB9YgSYh3itW': {
            'symbol': 'USDC',
            'name': 'Wrapped USDC',
        },
        'So11111111111111111111111111111111111111112': {
            'symbol': 'SOL',
            'name': 'Wrapped SOL',
        },
    }

    def __init__(self, address, *args, **kwargs):
        super().__init__(address)
        self._headers = {'Content-Type': 'application/json'}

    def get_balance(self):
        balances = []

        sol_balance = self._get_sol_balance()
        if sol_balance:
            balances.append(sol_balance)

        token_balances = self._get_token_balances()
        if token_balances:
            balances += token_balances

        return balances

    def _get_sol_balance(self):
        response = self._request(method='getBalance', params=[self.address])

        return {
            'symbol': self.symbol,
            'amount': to_decimal(response['result']['value']) * self.coef,
        }

    def _get_token_balances(self):
        response = self._request(
            method='getTokenAccountsByOwner',
            params=[
                self.address,
                {'programId': self.token_program_id},
                {'encoding': 'jsonParsed'},
            ],
        )

        balances = [self._parse_token_balance(b) for b in response['result']['value']]

        return [b for b in balances if b['amount'] > Decimal(0)]

    def _parse_token_balance(self, raw):
        info = raw['account']['data']['parsed']['info']

        token_address = info['mint']
        token_info = self.tokens.get(token_address)

        a = info['tokenAmount']
        if a.get('uiAmount'):
            amount = to_decimal(a['uiAmount'])
        elif a.get('decimals'):
            amount = to_decimal(a['amount']) * Decimal(f"10e-{a['decimals']}")
        else:
            amount = to_decimal(a['amount'])

        return {
            'symbol': token_info['symbol'] if token_info else 'UNKNOWN',
            'address': token_address,
            'amount': amount,
            'name': token_info['name'] if token_info else 'Unknown',
            'account_address': raw['pubkey'],
        }

    def get_txs_signatures(self, limit=None):
        if not limit:
            limit = self.max_items_per_page

        response = self._request(
            method='getConfirmedSignaturesForAddress2',
            params=[self.address, {'limit': limit}],
        )

        return [
            {'signature': r['signature'], 'error': bool(r['err'])}
            for r in response['result']
        ]

    def get_tx(self, signature):
        response = self._request(
            method='getConfirmedTransaction', params=[signature, 'jsonParsed']
        )
        return response['result']

    def _request(self, method, params):
        body = json.dumps(
            {'jsonrpc': '2.0', 'id': 1, 'method': method, 'params': params}
        )

        response = self.request(
            # request method is not needed, it's included in body
            request_method=None,
            body=body,
            headers=self._headers,
        )
        if response.get('error'):
            raise APIError(response['error'])

        return response



from blockapi.v2.coins import (
    COIN_ASTR,
    COIN_AURORA,
    COIN_AVAX,
    COIN_BNB,
    COIN_BOBA,
    COIN_BTT,
    COIN_CELO,
    COIN_CRO,
    COIN_ETH,
    COIN_FTM,
    COIN_FUSE,
    COIN_GLMR,
    COIN_IOTX,
    COIN_KCS,
    COIN_KLAY,
    COIN_MATIC,
    COIN_METIS,
    COIN_MOVR,
    COIN_OKT,
    COIN_ONE,
    COIN_OP,
    COIN_RSK,
    COIN_SGB,
    COIN_TLOS,
    COIN_WAN,
    COIN_XDAI,
)
from blockapi.v2.models import AssetType, Blockchain, Coin

DEBANK_ASSET_TYPES = {
    'vested': AssetType.VESTING,
    'liquidity pool': AssetType.LIQUIDITY_POOL,
}

REWARD_ASSET_TYPE_MAP = {
    AssetType.LENDING: AssetType.LENDING_REWARD,
    AssetType.STAKED: AssetType.REWARDS,
    AssetType.FARMING: AssetType.REWARDS,
    AssetType.YIELD: AssetType.REWARDS,
}


NATIVE_COIN_MAP = {
    'astar': COIN_ASTR,
    'avax': COIN_AVAX,
    'aurora': COIN_AURORA,
    'bsc': COIN_BNB,
    'boba': COIN_BOBA,
    'btt': COIN_BTT,
    'celo': COIN_CELO,
    'cro': COIN_CRO,
    'eth': COIN_ETH,
    'ftm': COIN_FTM,
    'fuse': COIN_FUSE,
    'hmy': COIN_ONE,
    'iotex': COIN_IOTX,
    'kcc': COIN_KCS,
    'klay': COIN_KLAY,
    'matic': COIN_MATIC,
    'metis': COIN_METIS,
    'mobm': COIN_GLMR,
    'movr': COIN_MOVR,
    'okt': COIN_OKT,
    'op': COIN_OP,
    'rsk': COIN_RSK,
    'sgb': COIN_SGB,
    'xdai': COIN_XDAI,
    'tlos': COIN_TLOS,
    'wan': COIN_WAN,
}


ALL_COINS = {
    var.symbol.lower(): var
    for var_name, var in globals().items()
    if isinstance(var, Coin) and var_name.startswith("COIN_")
}



from typing import Type

from blockapi import get_active_api_classes
from blockapi.services import BlockchainAPI
from blockapi.test_data import get_test_api_key, test_addresses


# TODO parametrize it for all endpoints using pytest
def _test_balances():
    for api_cls in get_active_api_classes():
        addr = test_addresses[api_cls.symbol][0]
        try:
            assert_balance(api_cls, addr)
        except AssertionError as e:
            print(e)


def assert_balance(api_cls: Type[BlockchainAPI], address: str):
    api_key = get_test_api_key(api_cls.__class__.__name__)
    api_inst = api_cls(address, api_key)

    balances = api_inst.get_balance()
    check_balance = next((b for b in balances if b['symbol'] == api_inst.symbol), None)

    assert check_balance, (
        f"No balance for symbol {api_inst.symbol} and API "
        f"{api_cls.__class__.__name__}."
    )



from datetime import datetime
from decimal import Decimal

import pytest

from blockapi.api.kyber import KyberAPI

USER_ADDRESS = '0x1375355fEcCB15e0DBedd7dcF3c496E2b1a25f6A'


class TestKyber:
    @staticmethod
    def test_raises_value_error():
        with pytest.raises(ValueError):
            KyberAPI(USER_ADDRESS, network='ropsten_typo')

    @staticmethod
    @pytest.mark.vcr()
    def test_get_staker_epoch_info():
        info = KyberAPI(USER_ADDRESS, network='ropsten').get_staker_epoch_info(182)
        assert list(info.keys()) == [
            'stake_amount',
            'delegated_stake_amount',
            'pending_stake_amount',
            'delegate',
        ]
        assert [type(val) for val in info.values()] == [Decimal, Decimal, Decimal, str]

    @staticmethod
    @pytest.mark.vcr()
    def test_get_actions():
        actions = KyberAPI(USER_ADDRESS, network='ropsten').get_user_actions()
        assert list(actions[0].keys()) == [
            'epoch',
            'type',
            'tx_hash',
            'meta',
            'action_date',
        ]
        assert [type(val) for val in actions[0].values()] == [
            Decimal,
            str,
            str,
            dict,
            datetime,
        ]

    @staticmethod
    @pytest.mark.vcr()
    def test_get_rewards():
        rewards = KyberAPI(USER_ADDRESS, network='ropsten').get_staker_rewards()
        assert list(rewards[0].keys()) == [
            'epoch',
            'amount',
            'claimed',
            'tx_hash',
            'total_stake',
            'total_reward',
            'total_voted',
        ]
        assert [type(val) for val in rewards[0].values()] == [
            Decimal,
            Decimal,
            bool,
            str,
            Decimal,
            Decimal,
            Decimal,
        ]

    @staticmethod
    @pytest.mark.vcr()
    def test_get_staker_votes():
        votes = KyberAPI(USER_ADDRESS, network='ropsten').get_staker_votes()
        assert list(votes[0].keys()) == [
            'staker',
            'epoch',
            'campaign_id',
            'option',
            'power',
        ]
        assert [type(val) for val in votes[0].values()] == [
            str,
            Decimal,
            Decimal,
            Decimal,
            str,
        ]



import pytest

from blockapi.v2.blockchain_mapping import (
    get_blockchain_from_chain_id,
    get_blockchain_from_coingecko_chain,
    get_blockchain_from_rango_chain,
)
from blockapi.v2.models import Blockchain


def test_map_case():
    assert get_blockchain_from_rango_chain('TERRA') == Blockchain.TERRA


def test_map_none():
    assert not get_blockchain_from_rango_chain(None)


@pytest.mark.parametrize(
    'chain,expected',
    [
        ('ShibChain', Blockchain.SHIBA_CHAIN),
        ('defi-kingdoms-blockchain', Blockchain.DEFI_KINGDOMS),
    ],
)
def test_map_coingecko(chain: str, expected: Blockchain):
    assert get_blockchain_from_coingecko_chain(chain) == expected


def test_map_by_id():
    assert get_blockchain_from_chain_id(1) == Blockchain.ETHEREUM
